/*
 *  WCOL: Access control (`check accept or not') routines.
 *      by k-chinen@is.aist-nara.ac.jp, 1994, 1995
 *
 *  $Id: accept.c,v 1.6 1995/10/15 18:39:35 k-chinen Exp k-chinen $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <signal.h>

#include "wcol.h"
#include "accept.h"



static int accept_last=0;
static accept_table_rec table[ACCEPT_TABLE_MAX];


/*
 * line must remove whites and not comment
 */
int AppendAcceptEntry(char *ltarget, char *lmask)
{
    unsigned char tmp[STRING_SIZE];
    unsigned char addr[5], mask[5];
    unsigned char *p,*q;
    int i;

    if(ltarget==NULL || !*ltarget
        || lmask==NULL || !*lmask) {
        Log("; 'AcceptAddress' configuration is wrong, please check syntax");
        return -1;
    }


    if(accept_last>=ACCEPT_TABLE_MAX) {
        Error("Read_Acceptable: Too many accept targets.");
        return -1;
    }


    /* Clear address and mask */
    memset(addr, 0, 5);
    memset(mask, 0, 5);



    /*
     * Get Address
     */
    p = (unsigned char*)ltarget;
    if(p==NULL || *p=='\0')
        return -1;
    i = 0;
    while(*p && i<4) {
        q = tmp;
        while(*p&&isdigit(*p))
            *q++ = *p++;
        *q = '\0';
        addr[i] = (unsigned char) atoi((char*)tmp);

        i++;
        if(*p=='.')
            p++;
        else
            break;
    }

    memcpy(table[accept_last].addr, addr, 5);
    Trace("AppendAcceptEntry: %d\n", accept_last);
    show_addr("\tAddress", addr);



    /*
     * Get Mask
     */
    p = (unsigned char*)lmask;
    if(p==NULL || *p=='\0') {
        goto default_mask;
    }

    i = 0;
    while(*p && i<4) {
        q = tmp;
        while(*p&&isdigit(*p))
            *q++ = *p++;
        *q = '\0';
        mask[i] = (unsigned char) atoi((char*)tmp);

        i++;
        if(*p=='.')
            p++;
        else
            break;
    }

    memcpy(table[accept_last].mask, mask, 5);
    show_addr("\tNetMask", mask);

    goto out;


    /*
     * Set Deafults Mask
     */

default_mask:
         if(addr[0]<128) {set_addr(mask, 255,   0,   0,   0);} /* A */
    else if(addr[0]<192) {set_addr(mask, 255, 255,   0,   0);} /* B */
    else if(addr[0]<224) {set_addr(mask, 255, 255, 255,   0);} /* C */
#if 0
    else if(addr[0]<240) {set_addr(mask, 255, 255, 255, 255);} /* D */
    else if(addr[0]<248) {set_addr(mask, 255, 255, 255, 255);} /* E */
#endif
    else {
        Trace("Ignore Address. realy ?\n");
    }

    memcpy(table[accept_last++].mask, mask, 5);
    show_addr("\tDefault NetMask", mask);


out:
    accept_last++;

    return 0;
}






/*
 * Acceptable - Check acceptable address or not.
 *
 *  return
 *      1   acceptable
 *      0   error
 *
 */
int Acceptable(unsigned char *target)
{
    int i,j;
    int match;

    if(accept_last==0)
        return 1;                   /* Because no access control */

    match = 0;
    for(i=0;i<accept_last;i++) {
        for(j=0;j<4;j++) {
            if((table[i].mask[j] & target[j]) != table[i].addr[j])
                break ;
        }

        if(j>=4) {
            match = 1;
#ifndef ACCEPT_TEST
            return 1;
#endif
        }
    }

    return match;
}



/*
 * Read_Accepttable - Read accept table.
 */
int Read_Accepttable(char *fname)
{
    FILE *fp;
    unsigned char line[STRING_SIZE], addr[STRING_SIZE], mask[STRING_SIZE];
    unsigned char *p,*q;

    accept_last = 0;
    if(fname==NULL || fname[0]=='\0') {
        Error("Read_Acceptable: no acceptable address file");
        return 0;
    }

    if((fp=fopen(fname, "r"))==NULL) {
        Error("Read_Acceptable: Cannot open '%s'", fname);

        Trace("Read_Acceptable: No Access Control\n");
        DLog("INFO no access control, file-'%s' is not found", fname);

        return 1;
    }

    Trace("Read accept table from %s ...\n", fname);
    while(fgets((char*)line, STRING_SIZE, fp)!=NULL) {
        p = line;
        /* Remove whites */
        while(*p && isspace(*p)) {
            p++;
        }

        /* store address parts */
        q = addr;
        while(*q && !isspace(*p)) {
            *q++ = *p++;
        }
        *q = '\0';

        /* skip whites */
        while(*p && isspace(*p)) {
            p++;
        }

        /* store mask parts */
        q = mask;
        while(*q && !isspace(*p)) {
            *q++ = *p++;
        }
        *q = '\0';

        AppendAcceptEntry((char*)addr,(char*) mask);
    }

    fclose(fp);

    if(accept_last==0)
        DLog("INFO no access control, file-'%s' is empty", fname);

    return 0;
}


#ifdef ACCEPT_TEST

/*
 *  Main routine for test.
 *
 *  If test to a1.a2.a3.a4 address is acceptable, run "$0 a1 a2 a3 a4".
 *
 */

extern char log_file[];
extern int trace_flag;


main(int argc, char **argv)
{
    int i;
    unsigned char target[5];

    trace_flag = 1;
    strcpy(log_file,"/dev/tty");

    Read_Accepttable("./wool-accept.cfg");

    i = 0;
    while(argv[i+1]) {
        target[i] = (unsigned char) atoi(argv[i+1]);
        i++;
    }

    show_addr("Ok, Target ", target);
    Trace(" -> %d\n",Acceptable(target));
}
#endif
