/*
 *  WCOL: path handle procedures
 *      by k-chinen@is.aist-nara.ac.jp, 1994, 1995, 1996, 1997
 *
 *  $Id: base.c,v 1.16 1996/11/24 14:47:50 k-chinen Exp k-chinen $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include <errno.h>

#include "wcol.h"
#include "url.h"


time_t cache_locktime=CACHELOCKTIME;

char pool_dir[STRING_SIZE];
int verbose = 0;


#ifdef DEBUG_BASE
int wcol_base_num_func_AssignFileName = 0;
#endif



int hostname_hash_radix=HOSTNAME_HASH_RADIX;

/*
 * Hash function for hostname
 */
static inline int hostname_hash(st)
char *st;
{
    register unsigned long k;
    register unsigned char *p;

    p = (unsigned char*) st;
    k = 0;
    while(*p) {
        k += *p++;
    }

    return (int)(k%hostname_hash_radix);
}




/*
 *  AssignFileName - assign URL to filename(path) in cache.
 *
 *  input:
 *      ret:    return address
 *      name:   URL
 *      cmd:    file type for assign (e.g., HEAD_FILE, HEAD_TMP_FILE, e.t.c.)
 *
 *  return:
 *      0:      success
 *      !0:     fail
 *
 *  Note:
 *      If you specify HEAD_TMP_FILE or BODY_TMP_FILE, this function
 *      return uniq filename by each call. Don't repeat, call once time.
 *
 */

int AssignFileName(char *ret, char *name, int cmd)
{
    static long seed=0;                     /* seed for uniq name */
    char protocol[STRING_SIZE];
    char host[STRING_SIZE], port[STRING_SIZE];
    char path[STRING_SIZE];
    char tmp[STRING_SIZE];
    char *ext;
    int  chk;
    register char *p,*q;

    int uniq_path(char*);


    if(name==NULL) {
        return  -1;
    }
    if(ret==NULL) {
        return  -2;
    }
    if(chk = ParseURL(name, protocol, host, port, path)) {
        return  -3;
    }

    seed++;

    /*
     * fix parameters
     */
    if(protocol[0]=='\0') {
        strcpy(protocol, "noprotocol");
    }
    if(host[0]=='\0') {
        strcpy(host, "nohost");
    }
    if(port[0]=='\0') {
        strcpy(port, "80");
    }
    if(path[0]=='\0') {
        strcpy(path, PADDING_NAME);
    }
    else {
        uniq_path(path);

        q = p = path;

        /* scan last charactor */
        p++;
        while(*p++)
            q++;
        if(*q=='/') { /* path is terminated with '/', padding dummy name */
            strcat(path, PADDING_NAME);
        }
    }

    switch(cmd) {
    case INFO_FILE:     ext = INFO_EXT;     break;
    case HIST_FILE:     ext = HIST_EXT;     break;
    case LOCK_FILE:     ext = LOCK_EXT;     break;
    case REQUEST_FILE:  ext = REQUEST_EXT;  break;
    case FACE_FILE:     ext = FACE_EXT;     break;
    case NOTE_FILE:     ext = NOTE_EXT;     break;

    case HEAD_FILE:     ext = HEAD_EXT;     break;
    case HEAD_TMP_FILE:
                        /* append uniq name */
                        sprintf(tmp, "%s-%05u",
                            HEAD_TMP_EXT, seed%100000);
                        ext = tmp;
                        break;
    case BODY_TMP_FILE:
                        /* append uniq name */
                        sprintf(tmp, "%s-%05u",
                            BODY_TMP_EXT, seed%100000);
                        ext = tmp;
                        break;
    default:
                        ext = "";
                        break;
    }

#if 0
    Trace("    protocol  '%s'\n", protocol);
    Trace("    hostname  '%s'\n", host);
    Trace("    hostname  %04d\n", hostname_hash(host));
    Trace("    port      '%s'\n", port);
    Trace("    path      '%s'\n", path);
    Trace("    ext       '%s'\n", ext);
#endif

    /*
     * protocol://host:port/path
     *  ->  pool-dir/protocol/host-hash/host/port/path,ext
     */
    sprintf(ret, "%s/%s/h%03d/%s/%s%s%s",
        pool_dir, protocol, hostname_hash(host), host, port, path, ext);

#if 0
    Trace("AssignFileName: path=%s\n    ->%s\n", name, ret);
#endif

    return 0;
}


int AssignURLToFileName(char *ret, url_t *url, int cmd)
{
    static long seed=0;                     /* seed for uniq name */
    char tmp[STRING_SIZE];
    char apath[STRING_SIZE];                /* actual path */
    char *ext;
    int  chk;
    register char *p,*q;


    if(chk = URL_Normalize(url)) {
        return  -3;
    }

    seed++;

#if 0
    Trace("AssignURLToFileName:\n");
    URL_Show(url);
#endif

    strcpy(apath, url->path);
    q = p = apath;

    /* scan last charactor */
    p++;
    while(*p++)
        q++;
    if(*q=='/') { /* path is terminated with '/', padding dummy name */
        strcat(apath, PADDING_NAME);
    }

    switch(cmd) {
    case INFO_FILE:     ext = INFO_EXT;     break;
    case HIST_FILE:     ext = HIST_EXT;     break;
    case LOCK_FILE:     ext = LOCK_EXT;     break;
    case REQUEST_FILE:  ext = REQUEST_EXT;  break;
    case FACE_FILE:     ext = FACE_EXT;     break;
    case NOTE_FILE:     ext = NOTE_EXT;     break;

    case HEAD_FILE:     ext = HEAD_EXT;     break;
    case HEAD_TMP_FILE:
                        /* append uniq name */
                        sprintf(tmp, "%s-%05u",
                            HEAD_TMP_EXT, seed%100000);
                        ext = tmp;
                        break;
    case BODY_TMP_FILE:
                        /* append uniq name */
                        sprintf(tmp, "%s-%05u",
                            BODY_TMP_EXT, seed%100000);
                        ext = tmp;
                        break;
    default:
                        ext = "";
                        break;
    }


    /*
     * protocol://host:port/path
     *  ->  pool-dir/protocol/host-hash/host/port/path,ext
     */
    sprintf(ret, "%s/%s/h%03d/%s/%s%s%s",
        pool_dir, url->proto, hostname_hash(url->host), url->host,
        url->port, apath, ext);

#if 0
    Trace("AssignURLToFileName: path=%s\n    ->%s\n", url->full_loc, ret);
#endif

    return 0;
}


/*
 *  AllocName - alloc pathname with URL
 *
 *  return      allocated path (it is exist in heap, please free it)
 *
 *  Note: If you specify HEAD_TMP_FILE or BODY_TMP_FILE, this function
 *      return uniq filename by each call. Don't repeat, call once time.
 */


/*
 *  AllocName - alloc assigned filename
 *
 *  input:
 *      ret:    return address
 *      name:   URL
 *      cmd:    file type for assign (e.g., HEAD_FILE, HEAD_TMP_FILE, e.t.c.)
 *
 *  return:
 *      NULL:   fail
 *      !NULL:  allocated path (it is exist in heap, please free it)
 *
 *  Note: If you specify HEAD_TMP_FILE or BODY_TMP_FILE, this function
 *      return uniq filename by each call. Don't repeat, call once time.
 */

char *AllocName(char *name, int cmd)
{
    char tmp[STRING_SIZE];
    char *ret;
    int  len;

    if(AssignFileName(tmp, name, cmd)) {
        return NULL;
    }

    return STRDUP(ret, tmp);
}



/*
 *
 */
int CheckFileType(char *path)
{
    int ret = UNKNOW_FILE;
    register char *p, *q;

    if(path==NULL)
        return -1;

    if(path[0]=='.' && (path[1]=='\0' || ((path[1]=='.'  && path[2]=='\0')))) {
        return -1;
    }

    p = path;
    q = path + strlen(path);
    --q;
    while(*q && p<q) {
        if(*q==',') {
            break;
        }
        --q;
    }

    /* unknow file */
    if(*q!=',') {
        ret = BODY_FILE;
    }
    else {
        if(strcmp(q, INFO_EXT)==0)      ret = INFO_FILE;
        if(strcmp(q, HIST_EXT)==0)      ret = HIST_FILE;
        if(strcmp(q, LOCK_EXT)==0)      ret = LOCK_FILE;
        if(strcmp(q, REQUEST_EXT)==0)   ret = REQUEST_FILE;
        if(strcmp(q, FACE_EXT)==0)      ret = FACE_FILE;
        if(strcmp(q, NOTE_EXT)==0)      ret = NOTE_FILE;

        if(strcmp(q, HEAD_EXT)==0)      ret = HEAD_FILE;
        if(strcmp(q, OLD_EXT)==0)       ret = UNKNOW_FILE;
        if(strncmp(q, HEAD_TMP_EXT, strlen(HEAD_TMP_EXT))==0)
            ret = HEAD_TMP_FILE;
        if(strncmp(q, BODY_TMP_EXT, strlen(BODY_TMP_EXT))==0)
            ret = BODY_TMP_FILE;

        /* /^.,ext$/ */
        if(q==p+1 && path[0]=='.') {
            return -1;
        }
        /* /^..,ext$/ */
        if(q==p+2 && path[0]=='.' && path[1]=='.' ) {
            return -1;
        }
    }

    return ret;
}






/*
 * ClearDir - clear dir (remove dirs and files). like 'rm -fr <dir>'
 *
 * return:
 *      -1  error
 *       0  success
 */
int ClearDir(char *dir)
{
    DIR         *dp;
    dirent_t    *elem;
    struct stat sbuf;
    char        path[STRING_SIZE];
    unsigned int n_dir, n_file, en_dir, en_file;

    static  long g_count=0;
    int count;
    int uniq_path(char *path);


    g_count++;
    count = g_count;

    Trace("ClearDir: start %4d '%s'\n", count, dir);
#if 0
    if(uniq_path(dir)) {
        Trace("     *ignore* path '%s'\n", dir);
        return -1;
    }
    else {
        Trace("     *adjust* path '%s'\n", dir);
    }
#endif

    n_dir = n_file = 0;
    en_dir = en_file = 0;
    if((dp = opendir(dir))==NULL) {
        Trace("ClearDir: *empty*\n");
        return -1;
    }
    while((elem=readdir(dp))!=NULL) {
        Trace("\tname '%s' ", elem->d_name);
        sprintf(path, "%s/%s", dir, elem->d_name);
        if(stat(path, &sbuf)) {
            Trace("error\n");
            continue;
        }
        if((elem->d_name[0]=='.' && elem->d_name[1]=='\0')
            || (elem->d_name[0]=='.' && elem->d_name[1]=='.'
                        && elem->d_name[2]=='\0')) {
            Trace("skip\n");
            continue;
        }

        if(S_ISDIR(sbuf.st_mode)) {
            Trace("\tdir\n");
            if(ClearDir(path)) {
                Trace("ClearDir: error remove dir  '%s'.\n", path);
                en_dir++;
            }
            else {
                n_dir++;
            }
        }
        else
        if(S_ISREG(sbuf.st_mode)) {
            Trace("\tfile ");
#if 1
            if(unlink(path)) {
                Trace("remove error\n");
                Error("ClearDir: error remove file '%s'", path);
                en_file++;
            }
            else {
                Trace("remove\n");
                n_file++;
            }
#else
            {
                Trace("remove(dummy)\n");
                n_file++;
            }
#endif
        }
        else
        {
            Trace("\tstrange-file\n", elem->d_name);

            Error("ClearDir: '%s' is not dir and reg-file.\n", path);
            closedir(dp);

            return -1;
        }
    }
    closedir(dp);

    Trace("ClearDir: check %4d '%s' dir %d, file %d, edir %d, efile %d\n",
         count, dir, n_dir, n_file, en_dir, en_file);

    if(en_dir + en_file != 0) {
        Trace("ClearDir: fail  %4d '%s'\n", count, dir);
        return -1;
    }

    rmdir(dir);

    Trace("ClearDir: done  %4d '%s'\n", count, dir);

    return 0;
}


/*
 * SetupDir - make hierarchical directories path
 *
 * return
 *      0   success
 *      1   error
 */
int SetupDir(char *path)
{
    char tmp[STRING_SIZE];
    struct stat sbuf;
    char *p;
    int ret=0;

#if 0
    Trace("SetupDir: 0 \"%s\"\n", path);
#endif

    /*
     * Loop parts of path separated by '/'
     */
    p = tmp;
    while(*path) {
        if(*path=='/') {
            *p++ = *path++;
        }
        while(*path&&*path!='/')  {
            *p++ = *path++;
        }
        *p = '\0';

        /*
        Trace("* SetupDir: check \"%s\".\r\n",tmp);
        */

        /* path is not terminated --- so, it is directory */
        if(*path) {
            /*
             * Not found it. So, create such directory.
             */
            if(stat(tmp,&sbuf)) {
                if(mkdir(tmp, 040755)<0 && errno!=EEXIST) {
                    Error("SetupDir: new dir,   Cannot mkdir '%s' errno %d",
                        tmp, errno);
                    ret = 1;
                    break;
                }
            }
            /*
             * Found it. Is it directory ?
             */
            else {
                /* found it, but it is directory */
                if(sbuf.st_mode & S_IFDIR) {
                    continue;
                }

                /* found it, but it is not directory */
                /*
                Trace("SetupDir: exist same path as file. cahnge to dir\n");
                */
                unlink(tmp);
                if(mkdir(tmp, 040755)<0 && errno!=EEXIST) {
                    Error("SetupDir: renew dir, Cannot mkdir '%s' errno %d",
                        tmp, errno);
                    ret = 1;
                    break;
                }
            }
        }
        /* path is terminated --- so, it is path (directory+filename) */
        else {
            break;
        }
    }

#if 0
    Trace("SetupDir: 2 return\n");
#endif

    return ret;
}



/*
 * SetupFile - setup file with SetupDir()
 *          open file which is named with naming scheme
 *
 * return
 *      0   success
 *      -1  error
 */
int SetupFile(char *nfname, char *name, int filetype, int *nfd)
{
    struct stat sbuf;

#if 0
    Trace("SetupFile: name '%s' type %d (%#x)\n", name, filetype, filetype);
#endif

    AssignFileName(nfname, name, filetype);
#if 0
    Trace("\tnfname '%s'\n", nfname);
#endif

    if(SetupDir(nfname)) {
        Error("SetupFile: Fail setup file \"%s\".", nfname);

        *nfd = -1;
        return -1;
    }

    if((*nfd=open(nfname,O_WRONLY|O_CREAT,0644))<0) {
        Error("SetupFile: Cannot open file (%d)", errno);
        Error("\tpath='%s'.", nfname);

        *nfd = -1;
        return -1;
    }

#if 0
    Trace("\t*nfd %d ->", *nfd);
    if(fstat(*nfd, &sbuf)<0) {
        Trace("fail fstat()\n");
    }
    else {
        Trace("mode %s, size %ld\n",
            S_ISREG(sbuf.st_mode) ? "REGULAR" : "not REGULAR",
            sbuf.st_size);
    }
#endif

    return 0;
}




/*
 * CheckLock - check 'Is it locked ?'
 *
 * return:
 *          0   not locked
 *          -1  locked
 */

int CheckLock(char *name)
{
    struct stat sbuf;
    char path[STRING_SIZE];
    int c;
    time_t now;

    c = 0;
    if(AssignFileName(path, name, LOCK_FILE)) {
        goto fail;
    }
    while(!stat(path,&sbuf)) {
        /*
         * locked long time, unlock
         */
        time(&now);
        if(now-sbuf.st_mtime>cache_locktime) {
/*
            Trace("CheckLock: %s was locked, But it is too old. remove !\n",
                name);
*/
            unlink(path);
            goto safe;
        }

        Trace("CheckLock: Locked (%dth try) %s\n\t%s\n",c,name,path);
        if(c>=LOCKRETRY_TIME) {
/*
            Trace("CheckLock: Locked %s\n\t%s\n",c,name,path);
*/
            goto fail;
        }

        sleep(LOCKRETRY_INTERVAL);
        c++;
    }

safe:
    FREE(path);

    return 0;

fail:
    FREE(path);

    return -1;
}





/*
 *  Lock - make LOCK-file.
 */
int Lock(char *name)
{
    char fname[STRING_SIZE];
    int fd;

    if(AssignFileName(fname, name, LOCK_FILE)) {
        return 1;
    }
    if(SetupDir(fname)) {

        Error("Lock: Cannot setup %s", fname);
        return 1;
    }
    if((fd=open(fname, O_WRONLY|O_CREAT, 0644))<0) {

        Error("Lock: Cannot lock %s", fname);
        return 1;
    }
    close(fd);

    return 0;
}

/*
 * UnkLock - remove LOCK-file.
 */
int UnLock(char *name)
{
    char fname[STRING_SIZE];

    if(!AssignFileName(fname, name, LOCK_FILE)) {
        unlink(fname);
    }

    return 0;
}



int RemoveExtension(char *path)
{
    register char *p, *q;

    p = path;
    q = path + strlen(path);
    q--;
    while(q!=p && *q!=',')
        q--;
    if(q!=p && *q==',') {
        *q = '\0';
        return 0;
    }
    return 1;
}





int ExtCmp(char *path, char *ext)
{
    register char *p;
    int ret=1;

    Trace("extcmp: enter '%s' vs '%s'\n",path,ext);

    p = path;
    while(*p)
        p++;
    while(p!=path && *p!='.')
        p--;
    if(*p=='.') {
        Trace("extcmp: try '%s' vs '%s'\n",p,ext);
        ret = strcmp(p,ext);
    }
    Trace("extcmp: diff. %s vs %s\n",p,ext);
    return ret;
}


/*
 *
 * return:
 *      1   HTML
 *      0   not HTML
 *
 */
int IsHTML(char *path)
{
    register char *p;
    int ret=0;

    p = path;
    while(*p)
        p++;
    if(p!=path&&*(p-1)=='/') {
        ret = 1;
        /* empty */
    }
    else {
        if(p!=path)
            p--;
        /*
         * seek period
         */
        while(p!=path&&*p!='.')
            p--;

        if(*p!='.' && *(p-1)=='/')  {
            ret = 1;
        }
        else if(*p=='.') {
            if(strcmp(p, PADDING_NAME)==0)  {
                ret = 1;
            }
            else if(strcasecmp(p, ".html")==0
                ||strcasecmp(p, ".htm")==0
                ||strcasecmp(p, ".shtml")==0)  {
                ret = 1;
            }
        }
    }

    return ret;
}



#if 0


/*
 *
 * return:
 *      0   not note or body
 *      1   note
 *      2   body
 *
 */
int IsURL_NOTEorBODY(char *name, char *realname)
{
    char *p, *q;

    Trace("IsURL_NOTEorBODY: hook to note, '%s'\n", name);

    p = q = name;
    while(*p)
        p++;

    p--;
    while(*p && q<p && *p != '*') {
        p--;
    }
    if(*p!='*') {
        goto not_note;
    }


    if(q<p && *(p-1)=='*') {
        p--;
        Trace("\tpiece '%s'\n", p);
        if(strcmp(p,"**NOTE")==0) {
            goto is_note;
        }
        if(strcmp(p,"**BODY")==0) {
            goto is_body;
        }
    }
    goto not_note;

is_note:
    strcpy(realname, name);
    q = realname + (p-name);
    *q = '\0';
    Trace("\tIt is NOTE '%s'\n", name);
    return 1;
    goto not_note;
is_body:
    strcpy(realname, name);
    q = realname + (p-name);
    *q = '\0';
    Trace("\tIt is BODY '%s'\n", name);

    return 2;
    goto not_note;

not_note:
    Trace("\tIt is not NOTE or BODY '%s'\n", name);
    return 0;

}

#endif


