/*
 * ctree.c -
 *      by k-chinen@is.aist-nara.ac.jp, 1997
 *
 * $Id: ctree.c,v 1.1 1997/09/15 01:58:48 k-chinen Exp k-chinen $
 */
/*
 * character tree
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <signal.h>
#include <setjmp.h>

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif /* HAVE_MALLOC_H */

#define MALLOC(x)   malloc(x)
#define FREE(x)     free(x)


#include "range.h"
#include "msg.h"
#include "sigs.h"
#include "ctree.h"


#define _TRACE_CTREE
#define _DEBUG_CTREE




ctree_t *ctree_new()
{
    ctree_t *ret;

    if((ret = (ctree_t*) MALLOC(sizeof(ctree_t)))==NULL) {
        Error("no memory");
        return NULL;
    }

    ret->label = '\0';
    ret->child = NULL;
    ret->brother = NULL;
    ret->value = NULL;
    ret->term  = 0;

#ifdef DEBUG_CTREE
    Trace("*** new %p\n", ret);
#endif

    return ret;
}


static int _ctree_regist(ctree_t *t, char *s, void *v)
{
#ifdef DEBUG_CTREE
    Trace("_ctree_regist: %p '%c'\n", t, *s);
#endif

    if(*s == '\0')  {
        t->term = 1;
        t->value = v;
#ifdef DEBUG_CTREE
        Trace("_ctree_regist: %p '%c' term    :%p %p\n",
            t, t->label, t->child, t->brother);
#endif
        return 0;
    }

    if(t->label == '\0') {
        t->label = *s;
    }

    if(t->label == *s) {
        if(t->child == NULL) {
            t->child = ctree_new();
        }
        _ctree_regist(t->child, s+1, v);
#ifdef DEBUG_CTREE
        Trace("_ctree_regist: %p '%c' child   :%p %p\n",
            t, t->label, t->child, t->brother);
#endif
    }
    else {
        if(t->brother == NULL) {
            t->brother = ctree_new();
        }

        _ctree_regist(t->brother, s, v);
#ifdef DEBUG_CTREE
        Trace("_ctree_regist: %p '%c' brother :%p %p\n",
            t, t->label, t->child, t->brother);
#endif
    }

    return 0;
}


int ctree_regist(ctree_t *t, char *s, void *v)
{
    int ret=1;

#ifdef TRACE_CTREE
    Trace("ctree_regist: base %p, label \"%s\" value %p\n", t, s, v);
#endif
    if(t==NULL) {
        return -1;
    }
    if(*s != '\0') {
        ret = _ctree_regist(t, s, v);
    }

    return ret;
}


int ctree_regist_case(ctree_t *t, char *ref, void *v)
{
    char xref[STRING_SIZE];
    register char *p, *q;

    if(ref) {
        p = ref;
        q = xref;
        while(*q++ = tolower(*p++))
            ;
        return ctree_regist(t, xref, v);
    }

    return -1;
}


/*****
 *****
 *****/

static char tree_label_stack[STRING_SIZE];
static int tree_label_stack_pos=0;
static int tree_item_count=0;

static int clear()
{
    tree_item_count=0;
    tree_label_stack_pos=0;
    return 0;
}

static int push(unsigned char ch)
{
    if(tree_label_stack_pos<(STRING_SIZE-1)) {
        tree_label_stack[tree_label_stack_pos] = ch;
        tree_label_stack_pos++;
        tree_label_stack[tree_label_stack_pos] = '\0';

#if 0
        Trace("-> <%s>\n", tree_label_stack);
#endif

        return 0;
    }
    else {
        Error("push: over run");
        return -1;
    }
}

static int pop()
{
    if(tree_label_stack_pos>0) {
        tree_label_stack_pos--;
        tree_label_stack[tree_label_stack_pos] = '\0';

#if 0
        Trace("<- <%s>\n", tree_label_stack);
#endif

        return 0;
    }
    else {
        Error("pop: under run");
        return -1;
    }
}



/*****
 *****
 *****/


static int _ctree_show_brother(ctree_t *t, FILE *outp);

static int _ctree_show_child(ctree_t *t, FILE *outp)
{
    if(t==NULL) {
        return 0;
    }


    if(t->term) {
        fprintf(outp, "%4d: %-40s %p\n",
            tree_item_count, tree_label_stack, t->value);
        fflush(outp);
        tree_item_count++;
    }


    if(t->child!=NULL) {
        push(t->label);
        _ctree_show_brother(t->child, outp);
        pop();
    }

    return 0;
}

static int _ctree_show_brother(ctree_t *t, FILE *outp)
{
    while(t) {
        _ctree_show_child(t, outp);
        t = t->brother;
    }

    return 0;
}

/*
int ctree_show(ctree_t *t, FILE *outp)
*/
int ctree_show(FILE *outp, char *prefix, ctree_t *t)
{
    int chk;

    clear();

    fprintf(outp, "; %s / addr %p \n", prefix, t);
    fprintf(outp, ";  #: %-40s %s\n", "label", "value");

    chk = _ctree_show_brother(t, outp);

    fprintf(outp, "; %d item(s) in this tree\n", tree_item_count);
    fflush(outp);

    return tree_item_count;
}


/*****
 *****
 *****/

static void* find_value;

static sigjmp_buf search_jmp_point;

static RETSIGTYPE jump_search_done()
{
    SIGLONGJMP(search_jmp_point, 1);
}


static int _ctree_search_brother(ctree_t *t, char *ref);

static int _ctree_search_child(ctree_t *t, char *ref)
{
    if(t==NULL) {
        return 0;
    }

    if(t->term) {
#ifdef DEBUG_CTREE
        Trace("[%s]\n", tree_label_stack);
#endif
        if(strcmp(tree_label_stack, ref)==0) {
            find_value = t->value;
            SIGLONGJMP(search_jmp_point, 1);
        }
    }


    if(t->child!=NULL) {
#ifdef DEBUG_CTREE
        Trace(" c <%c><%c>\n", t->label, ref[tree_label_stack_pos]);
#endif
        if(t->label == ref[tree_label_stack_pos]) {
            push(t->label);
            _ctree_search_brother(t->child, ref);
            pop();
        }
    }

    return 0;
}

static int _ctree_search_brother(ctree_t *t, char *ref)
{
    while(t) {
        _ctree_search_child(t, ref);
        t = t->brother;
    }

    return 0;
}

void *ctree_search(ctree_t *t, char *ref)
{
    int ret;

    clear();
    find_value = NULL;

#ifdef TRACE_CTREE
    Trace("ctree_search: addr %p, target string \"%s\"\n", t, ref);
#endif

    if(SIGSETJMP(search_jmp_point, 1)) {
#ifdef TRACE_CTREE
        Trace("  found at %p\n", find_value);
#endif
        return find_value; /* found */
    }

    _ctree_search_brother(t, ref);

#ifdef TRACE_CTREE
    Trace("  not found\n");
#endif

    return NULL; /* not found */
}

void *ctree_search_case(ctree_t *t, char *ref)
{
    char xref[STRING_SIZE];
    register char *p, *q;

    if(ref) {
        p = ref;
        q = xref;
        while(*q++ = tolower(*p++))
            ;

        return ctree_search(t, xref);
    }

    return NULL;
}







#ifdef TEST_CTREE

char dummy[STRING_SIZE] = "it is dummy";

void* EMPTY=dummy;


main()
{
    ctree_t   *greek;
    void* chk;

    trace_flag = 1;


    greek = ctree_new();
    ctree_show(stdout, "before", greek);

    ctree_regist(greek, "a", EMPTY);
    ctree_regist(greek, "b", EMPTY);
    ctree_regist(greek, "aK", EMPTY);
    ctree_regist(greek, "beta", EMPTY);
    ctree_regist(greek, "alpha", EMPTY);
    ctree_regist(greek, "attack", EMPTY);
    ctree_regist(greek, "alto", EMPTY);
    ctree_regist(greek, "a", EMPTY);
    ctree_regist(greek, "ah", EMPTY);
    ctree_regist(greek, "al", EMPTY);

    ctree_show(stdout, "after ", greek);

    chk = ctree_search(greek, "attacP");
    if(chk) {
        Trace("found\n");
    }
    else {
        Trace("not found\n");
    }
    chk = ctree_search(greek, "attack");
    if(chk) {
        Trace("found\n");
    }
    else {
        Trace("not found\n");
    }
}

#endif

