/*
 * icp.c - Internet Cache Protocol procedures
 *      by k-chinen@is.aist-nara.ac.jp, 1996, 1997
 *
 * $Id: icp.c,v 1.2 1997/07/03 12:54:11 k-chinen Exp k-chinen $
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <setjmp.h>
#include <signal.h>
#include <errno.h>


#include "wcol.h"
#include "sigs.h"
#include "socket.h"
#include "icp.h"
#include "msg.h"

#include "accept.h"

#define _ICP_TEST




#ifdef USE_ICP_NOTIFY

char *_icp_opcode_str[] = {
    "INVALID",
    "QUERY", "HIT", "MISS", "ERROR", "SEND", "SENDAUTH",
    "DATABEGIN", "DATA", "DATAEND", "SOURCEECHO", "DUMBCACHEECHO",
    "NOTIFY", "UNUSED1(NOTIFYHIT?)", "UNUSED2(NOTIFYIMS?)",
    "UNUSED3", "UNUSED4", "UNUSED5", "UNUSED6", "UNUSED7", "UNUSED8",
    "RELOADING", "DENIED", "HITOBJECT", "END",
};

#else

char *_icp_opcode_str[] = {
    "INVALID",
    "QUERY", "HIT", "MISS", "ERROR", "SEND", "SENDAUTH",
    "DATABEGIN", "DATA", "DATAEND", "SOURCEECHO", "DUMBCACHEECHO",
    "UNUSED0", "UNUSED1", "UNUSED2",
    "UNUSED3", "UNUSED4", "UNUSED5", "UNUSED6", "UNUSED7", "UNUSED8",
    "RELOADING", "DENIED", "HITOBJECT", "END",
};

#endif

static char *hex_str="0123456789ABCDEF?";


int ICP_IsHit(unsigned char *buf)
{
    return (((icp_message_t*)buf)->header.opcode == ICP_OP_HIT);
}
int ICP_IsMiss(unsigned char *buf)
{
    return (((icp_message_t*)buf)->header.opcode == ICP_OP_MISS);
}


void ICP_DumpMessage(FILE *output, unsigned char *buf)
{
    icp_message_t   *msg;
    unsigned char   *p, *q, *r;
    int             b,i;
    int             len;
    unsigned char   line[STRING_SIZE], tmp[STRING_SIZE];

    msg = (icp_message_t*) buf;
    len = ntohs(msg->header.length);

    fprintf(output,
"; op %-8s (%2d), ver %d, len %4d, req# %5d, flg %3d, pad %3d, host %d\n",
        ICPOperationStr(msg->header.opcode), msg->header.opcode,
        msg->header.version, ntohs(msg->header.length),
        ntohl(msg->header.reqnum),
        ntohl(msg->header.flags), msg->header.pad,
        ntohl(msg->header.shostid));

    p = buf;
    line[0] = '\0';
    i = 0;
    while(i<ICP_COMMON_SZ) {
        sprintf((char*)tmp, "%02x ", *p);
        p++;
        i++;

        strcat((char*)line, (char*)tmp);
    }
    fprintf(output, "; header: %s\n", line);

    while(i<len) {
        b = i;

        memset(line, ' ', STRING_SIZE);
        line[16*3+0] = '|';
        line[16*3+2+16*1] = '\0';

        q = &line[0];
        r = &line[16*3+2];
        do {
            *q++ = hex_str[*p/16];
            *q++ = hex_str[*p%16];
            *q++ = ' ';

            if(*p<' '||*p>0x7f)
                *r++ = '.';
            else
                *r++ = *p;

            p++;
            i++;
        } while((i-ICP_COMMON_SZ)%16>0 && i<len);

        fprintf(output, "; %06x: %s\n", b-ICP_COMMON_SZ, line);
    }

    fflush(output);
}


/*
 * procedure for QUERY request
 */
int ICP_SendQuery(int xfd, int rnum, unsigned char *target)
{
    unsigned char buf[KB(10)];
    icp_common_t *header;
    int  c;
    int  nwrite;
    int  len;

    /*
    Trace("ICP_SendQuery: target '%s'\n", target);
    */

    header = (icp_common_t*) buf;
    memset(header, 0, ICP_COMMON_SZ);
    header->opcode  = ICP_OP_QUERY;
    header->version = 2;
    header->reqnum  = rnum;
    {
        register unsigned char *p, *q;

        q = &buf[ICP_COMMON_SZ];
        *q++ = '\0';
        *q++ = '\0';
        *q++ = '\0';
        *q++ = '\0';
        p = target;
        c = 1;
        while(*q++=*p++)
            c++;
    }
    len = ICP_COMMON_SZ + 4 + c;
    header->length = htons(len);
    /*
    Trace("Data length %d (%d+%d)\n", len, ICP_COMMON_SZ, c);
    */

#if 1
    if(trace_flag) {
        ICP_DumpMessage(stdout, buf);
    }
#endif

    if((nwrite = send(xfd, (char*)buf, len, 0))<0) {
        Error("ICP_SendQuery: fail send() (%d)", errno);
        return -1;
    }
    /*
    Trace("ICP_SendQuery: success sendto() %d byte(s).\n", len);
    */

    return nwrite;
}

int ICP_SendToQuery(int xfd, struct sockaddr_in *from, int from_len,
    int rnum, unsigned char *target)
{
    unsigned char buf[KB(10)];
    icp_common_t *header;
    int  c;
    int  nwrite;
    int  len;

    /*
    Trace("ICP_SendToQuery: target '%s'\n", target);
    */

    header = (icp_common_t*) buf;
    memset(header, 0, ICP_COMMON_SZ);
    header->opcode  = ICP_OP_QUERY;
    header->version = 2;
    header->reqnum  = rnum;
    {
        register unsigned char *p, *q;

        q = &buf[ICP_COMMON_SZ];
        *q++ = '\0';
        *q++ = '\0';
        *q++ = '\0';
        *q++ = '\0';
        p = target;
        c = 1;
        while(*q++=*p++)
            c++;
    }
    len = ICP_COMMON_SZ + 4 + c;
    header->length = htons(len);
    /*
    Trace("Data length %d (%d+%d)\n", len, ICP_COMMON_SZ, c);
    */

#if 1
    ICP_DumpMessage(stdout, buf);
#endif

#if 0
fflush(stdout); fflush(stderr);
Trace("@@@ before sendto()\n");
#endif

    if((nwrite =
            sendto(xfd, (char*)buf, len, 0,
                (struct sockaddr*)from, sizeof(struct sockaddr_in)))<0) {
        Error("ICP_SendToQuery: fail sendto() (%d)", errno);
        return -1;
    }

#if 0
Trace("@@@ after  sendto()\n");
fflush(stdout); fflush(stderr);
#endif

    /*
    Trace("ICP_SendToQuery: success sendto() %d/%d byte(s).\n", nwrite, len);
    */

    return nwrite;
}


/*
 * procedure for QUERY request
 */
int ICP_SendNotify(int xfd, int rnum, unsigned char *target)
{
    unsigned char buf[KB(10)];
    icp_common_t *header;
    int  c;
    int  nwrite;
    int  len;

    /*
    Trace("ICP_SendNotify: target '%s'\n", target);
    */

    header = (icp_common_t*) buf;
    memset(header, 0, ICP_COMMON_SZ);
    header->opcode  = ICP_OP_NOTIFY;
    header->version = 2;
    header->reqnum  = rnum;
    {
        register unsigned char *p, *q;

        q = &buf[ICP_COMMON_SZ];
        *q++ = '\0';
        *q++ = '\0';
        *q++ = '\0';
        *q++ = '\0';
        p = target;
        c = 1;
        while(*q++=*p++)
            c++;
    }
    len = ICP_COMMON_SZ + 4 + c;
    header->length = htons(len);
    /*
    Trace("Data length %d (%d+%d)\n", len, ICP_COMMON_SZ, c);
    */

#if 1
    if(trace_flag) {
        ICP_DumpMessage(stdout, buf);
    }
#endif

    if((nwrite = send(xfd, (char*)buf, len, 0))<0) {
        Error("ICP_SendNotify: fail send() (%d)", errno);
        return -1;
    }
    /*
    Trace("ICP_SendNotify: success sendto() %d byte(s).\n", len);
    */

    return nwrite;
}

int ICP_SendToNotify(int xfd, struct sockaddr_in *from, int from_len,
    int rnum, unsigned char *target)
{
    unsigned char buf[KB(10)];
    icp_common_t *header;
    int  c;
    int  nwrite;
    int  len;

    /*
    Trace("ICP_SendToNotify: target '%s'\n", target);
    */

    header = (icp_common_t*) buf;
    memset(header, 0, ICP_COMMON_SZ);
    header->opcode  = ICP_OP_NOTIFY;
    header->version = 2;
    header->reqnum  = rnum;
    {
        register unsigned char *p, *q;

        q = &buf[ICP_COMMON_SZ];
        *q++ = '\0';
        *q++ = '\0';
        *q++ = '\0';
        *q++ = '\0';
        p = target;
        c = 1;
        while(*q++=*p++)
            c++;
    }
    len = ICP_COMMON_SZ + 4 + c;
    header->length = htons(len);
    /*
    Trace("Data length %d (%d+%d)\n", len, ICP_COMMON_SZ, c);
    */

#if 1
    ICP_DumpMessage(stdout, buf);
#endif

#if 0
fflush(stdout); fflush(stderr);
Trace("@@@ before sendto()\n");
#endif

    if((nwrite =
            sendto(xfd, (char*)buf, len, 0,
                (struct sockaddr*)from, sizeof(struct sockaddr_in)))<0) {
        Error("ICP_SendToNotify: fail sendto() (%d)", errno);
        return -1;
    }

#if 0
Trace("@@@ after  sendto()\n");
fflush(stdout); fflush(stderr);
#endif

    /*
    Trace("ICP_SendToNotify: success sendto() %d byte(s).\n", len);
    */

    return nwrite;
}


/*
 * procedure for MISS reply
 */
int ICP_SendMiss(int xfd, struct sockaddr_in *from, int from_len,
    icp_common_t *bheader, unsigned char *target)
{
    unsigned char buf[KB(10)];
    icp_common_t *header;
    int  c;
    int  nwrite;
    int  len;

    /*
    Trace("ICP_SendMiss: target '%s'\n", target);
    */

    header = (icp_common_t*) buf;
    memcpy(header, bheader, ICP_COMMON_SZ);
    header->opcode = ICP_OP_MISS;
    {
        register unsigned char *p, *q;

        q = &buf[ICP_COMMON_SZ];
        p = target;
        c = 1;
        while(*q++=*p++)
            c++;
    }
    len = ICP_COMMON_SZ + c;
    header->length = htons(len);
    /*
    Trace("Data length %d (%d+%d)\n", len, ICP_COMMON_SZ, c);
    */

#if 1
    fflush(stdout);
    ICP_DumpMessage(stdout, buf);
#endif

    if((nwrite =
            sendto(xfd, (char*)buf, len, 0,
                (struct sockaddr*)from, sizeof(struct sockaddr_in)))<0) {
        Error("ICP_SendMiss: fail sendto() (%d)", errno);
        return -1;
    }

    /*
    Trace("ICP_SendMiss: success sendto() %d byte(s).\n", len);
    */

    return nwrite;
}

/*
 * procedure for HIT reply
 */
int ICP_SendHit(int xfd, struct sockaddr_in *from, int from_len,
    icp_common_t *bheader, unsigned char *target)
{
    unsigned char buf[KB(10)];
    icp_common_t *header;
    int  c;
    int  nwrite;
    int  len;

    /*
    Trace("ICP_SendHit: target '%s'\n", target);
    */

    header = (icp_common_t*) buf;
    memcpy(header, bheader, ICP_COMMON_SZ);
    header->opcode = ICP_OP_HIT;
    {
        register unsigned char *p, *q;

        q = &buf[ICP_COMMON_SZ];
        p = target;
        c = 1;
        while(*q++=*p++)
            c++;
    }
    len = ICP_COMMON_SZ + c;
    header->length = htons(len);
    /*
    Trace("Data length %d (%d+%d)\n", len, ICP_COMMON_SZ, c);
    */

#if 1
    fflush(stdout);
    ICP_DumpMessage(stdout, buf);
#endif

    if((nwrite =
            sendto(xfd, (char*)buf, len, 0,
                (struct sockaddr*)from, sizeof(struct sockaddr_in)))<0) {
        Error("ICP_SendHit fail sendto() (%d)", errno);
        return -1;
    }

    /*
    Trace("ICP_SendHit success sendto() %d byte(s).\n", len);
    */

    return nwrite;
}



#ifdef ICP_TEST


/*
 * NOTE:    This version reply 'HIT' to HTTP objects.
 */






#if 0
int xICP_Request(char *hostname, int portno, char *url)
{
    int netd;
    struct sockaddr_in addr;
    static int count=0;

    if((netd = UDP_connect(hostname, portno, &addr))<0) {
        Error("ICP_Request: cannot setup UDP socket");
    }

#if 1
    ICP_SendToQuery(netd, &addr, sizeof(addr), ++count, url);
#else
    ICP_SendQuery(netd, ++count, url);
#endif
}
#endif

/*
 *  return:
 *      -1  fail to connect
 *      0   success
 *      1   no reply
 *
 */
int ICP_Request(char *hostname, int portno, char *url)
{
    int netd;
    char buf[BUF_SIZE];
    int nread;
    static int count=0;

    if((netd = UDP_connect(hostname, portno))<0) {
        Error("ICP_Request: cannot setup UDP socket");
        return -1;
    }

    ICP_SendQuery(netd, ++count, (unsigned char*)url);
    /*
    fflush(stderr);
    */
    fflush(stdout);

    if((nread = read(netd, buf, BUF_SIZE))<0) {
        Trace("ICP_Request: No reply\n");
        return 1;
    }

    if(trace_flag) {
        ICP_DumpMessage(stdout, (unsigned char*)buf);
        fflush(stdout);
    }

    return 0;
}






/*
 *
 */
reply_process(int listen_port)
{
    struct sockaddr_in from;
    int from_len;
    int icp_fd;
    int nread;
    unsigned char buf[KB(10)];
    icp_common_t *header;
    unsigned char *target;
    unsigned char *peer_addr;



    Trace("size of icp header common part = %d\n", ICP_COMMON_SZ);
    Trace("size of icp header = %d\n", ICP_HDR_SZ);

    if((icp_fd = UDP_bind_port(listen_port, WCOL_LISTEN_LENGTH)) < 0) {
        Error("icp: cannot bind listen port");
        exit(0);
    }
    Trace("icp: port is ready\n");


try0:
    while(from_len = sizeof(struct sockaddr),
            (nread = recvfrom(icp_fd, buf, sizeof(buf), 0,
                    (struct sockaddr*)&from, &from_len))>=0) {
        if(nread == 0) {
            Trace("\nicp: Empty\n");
        }
        else {
            Trace("\nicp: Read %d bytes\n", nread);
            Trace("icp: From length %d ", from_len);
            peer_addr = (unsigned char*) &from.sin_addr.s_addr;
            show_addr(", addr ", peer_addr);

#if 1
            ICP_DumpMessage(stdout, buf);
#endif
            header = (icp_common_t*) buf;
            if(header->opcode == ICP_OP_QUERY) {
                target = &buf[ICP_HDR_SZ];
                Trace("\tQuery URL = '%s'\n", target);

                if(strncmp(target,"http:",5)==0) {
                    ICP_SendHit(icp_fd, &from, from_len, header, target);
                }
                else {
                    ICP_SendMiss(icp_fd, &from, from_len, header, target);
                }

            }
            else {
                Trace("Unsupport ICP message\n");
#if 1
                ICP_DumpMessage(stdout, buf);
#endif
            }
        }
    }
    if(nread<0 && errno == EINTR)
        goto try0;

    Trace("icp: complete.");


}


query_usage()
{
    fprintf(stderr, "icp's query mode, please check arguments\n\n");
    fprintf(stderr, "usage: icp -q host port# url\n\n");
}


sigjmp_buf icp_query_jumpbuf;

static RETSIGTYPE giveup_icp_query()
{
    SIGLONGJMP(icp_query_jumpbuf, 1);
}


/*
 *
 */
query_process(char *qhost, char *qport, char *qurl, int try_limit)
{
    int try_num, ok, chk;

    /*
     * query loop
     */
    fprintf(stdout, "Query: host '%s', port '%s' (%d), URL '%s'\n",
        qhost, qport, atoi(qport), qurl);
    fprintf(stdout, "Try %d times\n", try_limit);
    fflush(stdout);

    try_num=0;
    ok = 0;
    while(try_num < try_limit && !ok) {
        try_num++;
        if(!trace_flag) {
            fprintf(stdout, "%05d/%05d ", try_num, try_limit);
        }

        if(SIGSETJMP(icp_query_jumpbuf, 1)) {
            if(trace_flag) {
                fprintf(stdout, "%05d/%05d th query *timeout*\n",
                    try_num, try_limit);
            }
            else {
                fprintf(stdout, "timeout\n");
            }
            fflush(stdout);
            continue;
        }

        SetSignalHandler(SIGALRM, giveup_icp_query);
        alarm(3);

        if(chk = ICP_Request(qhost, atoi(qport), qurl)) {
            if(trace_flag) {
                fprintf(stdout, "%05d/%05d th query *fail* (%d)\n",
                    try_num, try_limit, chk);
            }
            else {
                fprintf(stdout, "*fail* (%d)\n", chk);
            }
            fflush(stdout);
            sleep(1);
        }
        else {
            if(!trace_flag)
                fprintf(stdout, "*ok*\n");

            ok = 1;
        }
    }

    /*
     * answer
     */
    if(ok) {
        fprintf(stdout, "target system is fine.\n");
        exit(0);
    }
    else {
        fprintf(stdout, "target system is strange. down or very busy.\n");
        exit(1);
    }
}


usage()
{
    fprintf(stderr,
"icp - ICP (Internet Cache Protocol) test program\n\
                by k-chinen@is.aist-nara.ac.jp, 1996\n\
\n\
usage:\n\
    icp -l [-p port#] [-n num]\n\
    icp [-n count#] -q host port# url\n\
    icp -h\n\
\n\
default paramters:\n\
    ICP port        3130\n\
    number of try     10\n\
\n\
example:\n\
    %% icp -l -p 3130\n\
    %% icp -q japan.park.org 3130 http://shika.aist-nara.ac.jp\n\
    %% icp -n 100 -q japan.park.org 3130 http://shika.aist-nara.ac.jp\n\
\n");

}



main(int argc, char **argv)
{
    int query_mode = 0, listen_mode = 0;
    int listen_port = ICP_PORTNO;

    int flag;
    extern char *optarg;
    extern int  optind;

    /* for query mode */
    char qhost[STRING_SIZE], qport[STRING_SIZE], qurl[STRING_SIZE];
    int try_limit;


    /* set defaults */
    try_limit = 10;
    trace_flag = 1;

    /* parse args */
    while((flag = getopt(argc, argv, "hp:tqln:")) != -1) {
        switch(flag) {
        case 'u':
            usage();
            exit(1);
            break;
        case 'p':
            listen_port = atoi(optarg);
            break;
        case 't':
            trace_flag = 1 - trace_flag;
            break;
        case 'q':
            query_mode = 1;
            break;
        case 'l':
            listen_mode = 1;
            break;
        case 'n':
            try_limit = atoi(optarg);
            break;
        }
    }

    if(query_mode) {
        if(argv[optind]) {
            strcpy(qhost, argv[optind++]);
        }
        else {
            query_usage();
            return;
        }
        if(argv[optind]) {
            strcpy(qport, argv[optind++]);
        }
        else {
            query_usage();
            return;
        }
        if(argv[optind]) {
            strcpy(qurl,  argv[optind++]);
        }
        else {
            query_usage();
            return;
        }

        query_process(qhost, qport, qurl, try_limit);
    }
    else
    if(listen_mode) {
        reply_process(listen_port);
    }
    else {
        usage();
    }

    exit(0);
}
#endif

