/*
 *  WCOL: Japanese Filter procedures
 *      by k-chinen@is.aist-nara.ac.jp, 1994, 1995
 *
 *  $Id: jfilter.c,v 1.6 1995/12/04 17:19:31 k-chinen Exp k-chinen $
 */

/*
 *  There are 3 major japanese code. These are called JIS, SJIS and EUC.
 *  In generally, People use JIS code (what is known as ISO-2022-JP) in
 *  HTML. But several OS cannot handle JIS, Then we need convertor.
 *  This procedures are a kind of answer for that problem.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#include <limits.h>
#include <ctype.h>

/*
 * If you want test, please compile with -DJFILETER_TEST.
 */
#ifdef JFILTER_TEST
#define BUF_SIZE 10240
#else
#include "wcol.h"
#endif

#include "jfilter.h"


#if 0

#ifndef JIS_CODE
#define JIS_CODE    (0)
#define EUC_CODE    (1)
#define SJIS_CODE   (2)
#endif /* !JIS_CODE */

#endif



#define iseuc(x)    ((x) & 0x80)
#define issjis1(x)  ((0x81 <= (x) && (x) <= 0x9f) \
                        || (0xe0 <= (x) && (x) <= 0xef))
#define issjis2(x)  ((0x40 <= (x) && (x) <= 0x7e) \
                        || (0x80 <= (x) && (x) <= 0xfc))

/* SJIS use C1-area, We can use it to identify */
#define ident_sjis1(x)  (0x81 <= (x) && (x) <= 0x9f)
#define ident_sjis2(x)  ((0x40 <= (x) && (x) <= 0x7e) \
                        || (0x80 <= (x) && (x) <= 0xfc))

/*
 * Core of conversion, It is very IMPORTANT.
 */

#define _jistosjis(h, l) \
    (((l += (h & 0x01 ? 0x1f: 0x7d)) >= 0x7f) ? (int)++l  : (int)l) + \
    ((((h = (h - 0x21 >> 1) + 0x81) >= 0x9f) ?  h += 0x40 : h) << 8)

#define _jistoeuc(h, l) \
    ((int)(l |= 0x80)<<8) + (int)(h |= 0x80)
#define _euctojis(h, l) \
    ((int)(l &= 0x7f)<<8) + (int)(h &= 0x7f)

#define _euctosjis(h, l) \
    (h &= 0x7f, l &= 0x7f, _jistosjis(h, l))

#if 0

/* for ULTRIX */
#define _sjistojis(h, l) \
    ( h = ((h -= (h < 0xa0) ? 0x71: 0xb1) << 1) + 1, \
        l > 0x7f ? l--:l, \
        l > 0x9d ? (++h,l -= 0x7d) : l -= 0x1f, \
            (h<<8)+l)
#define _sjistoeuc(h, l) \
    ( _sjistojis(h,l), (h |= 0x80)<<8 + (l |= 0x80) )

#else

/* for BSD/OS and others */
#define _sjistojis(h, l) { \
        h = ((h -= (h < 0xa0) ? 0x71: 0xb1) << 1) + 1; \
        if(l>0x7f) l--; \
        if(l>0x9d) {++h; l -= 0x7d;} else {l -= 0x1f;} \
    }
#define _sjistoeuc(h, l) \
    { _sjistojis(h,l); h |= 0x80; l |= 0x80; }

#endif








/*
 * CheckKanjiCode - Check kanji code of charactors in buffer.
 *      For japanese (JIS/SJIS/EUC-code) documents and ASCII only. Don't
 *      apply other japanese code documents or other language documents.
 *      Because this procedure recognize escaped string as JIS-code.
 *      If you apply this procedure to other language document, all
 *      escaped strings recognized to JIS-code.
 *
 * return
 *      -1 unknown ASCII or others
 *       0 JIS (ISO-2022,7bit) - escaped string
 *       1 SJIS
 *       2 EUC
 */
int CheckKanjiCode(char *buf)
{
    register unsigned char *p;
    int ret, eightbit;

    ret = -1;
    eightbit = 0;
    p = (unsigned char*) buf;
    while(*p) {
        /* skip ASCII */
        while(*p && isascii(*p)) {
            if(*p == 0x1b) /* ESCAPE */
                return JIS_CODE;    /* JIS */
            p++;
        }

        /* not ASCII */
        if(*p) {
            eightbit = 1;
            /* SJIS check; first 0x81-0x9f, second 0x40-0x7e,0x80-0xfc */
            if(0x81 <= *p && *p <= 0x9f) {
                p++;
                if((0x40 <= *p && *p <= 0x7e)
                    || (0x80 <= *p && *p <= 0xfc)) {
                        return SJIS_CODE; /* SJIS */
                }
            }
            /* EUC or SJIS */
            else {
                p++;
            }

            if(*p)
                p++;
        }
    }

    if(eightbit) {
        return EUC_CODE; /* EUC */
    }
    else {
        return -1;
    }
}


/*
 * CheckFileKanjiCode - Check kanji code in file.
 *
 */
int CheckFileKanjiCode(char *filename, int linelimit)
{
    FILE *fp;
    char buf[512];
    int  line;
    int  code;
    int  j,sj,e;

    if((fp=fopen(filename,"r"))==NULL) {
        return -1;
    }

    j = sj = e = 0;
    line = 0;
    while(fgets(buf, 512, fp)!=NULL && line<linelimit) {
        code = CheckKanjiCode(buf);
#ifdef DEBUG
        printf(" %2d %d (%s)", line, code, buf);
#endif
        line++;
        if(code>=0) {
            switch(code) {
            case JIS_CODE:  /* JIS */
                j++;
                break;
            case SJIS_CODE: /* SJIS */
                sj++;
                break;
            case EUC_CODE:  /* EUC */
                e++;
                break;
            }
        }
    }
    fclose(fp);

    if(sj)
        return SJIS_CODE;
    if(e)
        return EUC_CODE;
    if(j)
        return JIS_CODE;

    return -1;
}




/*
 * copy as is
 */
static int thru(int ifd, int ofd)
{
    unsigned char ibuffer[BUF_SIZE+1];
    int len;

    while((len=read(ifd, ibuffer, BUF_SIZE))>0) {
        write(ofd, ibuffer, len);
    }

    return 0;
}



/*
 *  jistosjis - convert JIS to SJIS.
 *
 *      It is dirty. Please to be beauty.
 *
 *      Flow:
 *          1. parse escape sequence
 *          2. change mode
 *          3. convert charactor to SJIS
 *
 *      Mode:
 *          JIS have many mode. This routine support 2 mode (KANJI and
 *          ASCII), 4 escape seqence (2 way of KANJI-in/out).
 *
 *      Level:
 *          Escape sequence is multi charactor. this routine count it as
 *          'level'.
 *
 *                  0 ----> 1 ----> 2 ----> <kanji mode>
 *                    ESC   |   $   |   @
 *                          |       |
 *                          |       +-----> <kanji mode>
 *                          |           B
 *                          |
 *                          +-----> 12 ---> <ascii mode>
 *                              (   |   B
 *                                  |
 *                                  +-----> <ascii mode>
 *                                      J
 *
 */
static int jistosjis(int ifd, int ofd)
{
    unsigned char ibuffer[BUF_SIZE+1], obuffer[BUF_SIZE];
    register unsigned char *p, h, l;
    register int level, flag, c;
    int len;

    level = 0;
    flag = 0;
    while((len=read(ifd, ibuffer, BUF_SIZE))>0) {
        ibuffer[len] = '\0';
        p = ibuffer;
        c = 0;
        while(*p) {
            if(*p==0x1b) {
                level = 1;
            }
            else {
                if(level) {
                    if(level==1) {
                        if(*p=='$')         level =  2;
                        else if(*p=='(')    level = 12;
                    }
                    else {
                        /* Translation */
                        if(level== 2 && *p=='@')    flag = 1;
                        if(level== 2 && *p=='B')    flag = 1;
                        if(level==12 && *p=='B')    flag = 0;
                        if(level==12 && *p=='J')    flag = 0;

                        /* Give up to parse escape sequence */
                        level=0;
                    }
                }
                else {
                    /* KANJI mode without control charactors */
                    if(flag && *p>=0x20) {
                        h = *p;
                        p++;
                        l = *p;
                        _jistosjis(h,l);
                        obuffer[c++] = h;
                        obuffer[c++] = l;
                    }
                    /* ASCII mode or control charactor in KANJI mode */
                    else {
                        obuffer[c++] = *p;
                    }
                }
            }
            p++;
        }
        write(ofd, obuffer, c);
    }
    return 0;
}

/*
 *  jistoeuc - convert JIS to EUC.
 *      This is same function jistosjis(), without it is conversion to EUC.
 */
static int jistoeuc(int ifd, int ofd)
{
    unsigned char ibuffer[BUF_SIZE+1], obuffer[BUF_SIZE];
    register unsigned char *p, h, l;
    register int level, flag, c;
    int len;

    level = 0;
    flag = 0;
    while((len=read(ifd, ibuffer, BUF_SIZE))>0) {
        ibuffer[len] = '\0';
        p = ibuffer;
        c = 0;
        while(*p) {
            if(*p==0x1b) {
                level = 1;
            }
            else {
                if(level) {
                    if(level==1) {
                        if(*p=='$')         level = 2;
                        else if(*p=='(')    level = 12;
                    }
                    else {
                        /* Translation */
                        if(level== 2 && *p=='@')    flag = 1;
                        if(level== 2 && *p=='B')    flag = 1;
                        if(level==12 && *p=='B')    flag = 0;
                        if(level==12 && *p=='J')    flag = 0;

                        /* Give up to parse escape sequence */
                        level=0;
                    }
                }
                else {
                    /* KANJI mode without control charactors */
                    if(flag && *p>=0x20) {
                        h = *p;
                        p++;
                        l = *p;
                        _jistoeuc(h,l);
                        obuffer[c++] = h;
                        obuffer[c++] = l;
                    }
                    /* ASCII mode or control charactor in KANJI mode */
                    else {
                        obuffer[c++] = *p;
                    }
                }
            }
            p++;
        }
        write(ofd, obuffer, c);
    }
    return 0;
}


static int euctosjis(int ifd, int ofd)
{
    unsigned char ibuffer[BUF_SIZE+1], obuffer[BUF_SIZE];
    register unsigned char *p, *q, h, l;
    register int c;
    int len;

    while((len=read(ifd, ibuffer, BUF_SIZE))>0) {
        ibuffer[len] = '\0';
        p = ibuffer;
        q = obuffer;
        c = 0;
        while(*p) {
            while(iseuc(*p) && iseuc(*(p+1))) {
                h = *p;
                l = *(p+1);
                _euctosjis(h, l);
                /*
                _euctojis(h, l);
                _jistosjis(h, l);
                */
                *q = h;
                *(q+1) = l;
                p += 2;
                q += 2;
                c += 2;
            }
            while(*p && !iseuc(*p)) {
                *q++ = *p++;
                c++;
            }
        }
        write(ofd, obuffer, c);
    }
    return 0;
}

static int euctojis(int ifd, int ofd)
{
    unsigned char ibuffer[BUF_SIZE+1], obuffer[BUF_SIZE*4];
    register unsigned char *p, *q, h, l;
    register int c;
    int len;

    while((len=read(ifd, ibuffer, BUF_SIZE))>0) {
        ibuffer[len] = '\0';
        p = ibuffer;
        q = obuffer;
        c = 0;
        while(*p) {
            if(iseuc(*p) && iseuc(*(p+1))) {
                *q++ = 0x1b;
                *q++ = '$';
                *q++ = 'B';
                c += 3;

                while(iseuc(*p) && iseuc(*(p+1))) {
                    h = *p;
                    l = *(p+1);
                    _euctojis(h, l);
                    *q = h;
                    *(q+1) = l;
                    p += 2;
                    q += 2;
                    c += 2;
                }

                *q++ = 0x1b;
                *q++ = '(';
                *q++ = 'B';
                c += 3;
            }
            while(*p && !iseuc(*p)) {
                *q++ = *p++;
                c++;
            }
        }
        write(ofd, obuffer, c);
    }
    return 0;
}




static int sjistojis(int ifd, int ofd)
{
    unsigned char ibuffer[BUF_SIZE+1], obuffer[BUF_SIZE*4];
    register unsigned char *p, *q, h, l;
    register int c;
    int len;

    while((len=read(ifd, ibuffer, BUF_SIZE))>0) {
        ibuffer[len] = '\0';
        p = ibuffer;
        q = obuffer;
        c = 0;
        while(*p) {
            if(issjis1(*p) && issjis2(*(p+1))) {
                *q++ = 0x1b;
                *q++ = '$';
                *q++ = 'B';
                c += 3;

                while(issjis1(*p) && issjis2(*(p+1))) {
                    h = *p;
                    l = *(p+1);
                    _sjistojis(h, l);
                    *q = h;
                    *(q+1) = l;
                    p += 2;
                    q += 2;
                    c += 2;
                }

                *q++ = 0x1b;
                *q++ = '(';
                *q++ = 'B';
                c += 3;
            }
            while(*p && !issjis1(*p)) {
                *q++ = *p++;
                c++;
            }
        }
        write(ofd, obuffer, c);
    }
    return 0;
}



static int sjistoeuc(int ifd, int ofd)
{
    unsigned char ibuffer[BUF_SIZE+1], obuffer[BUF_SIZE*4];
    register unsigned char *p, *q, h, l;
    register int c;
    int len;

    while((len=read(ifd, ibuffer, BUF_SIZE))>0) {
        ibuffer[len] = '\0';
        p = ibuffer;
        q = obuffer;
        c = 0;
        while(*p) {
            while(issjis1(*p) && issjis2(*(p+1))) {
                h = *p;
                l = *(p+1);
                _sjistoeuc(h, l);
                /*
                _sjistojis(h, l);
                h |= 0x80;
                l |= 0x80;
                */
                *q = h;
                *(q+1) = l;
                p += 2;
                q += 2;
                c += 2;
            }
            while(*p && !issjis1(*p)) {
                *q++ = *p++;
                c++;
            }
        }
        write(ofd, obuffer, c);
    }
    return 0;
}




/*
 */
int ToEUC(int icode, int ifd, int ofd)
{
    switch(icode) {
    case JIS_CODE:  /* JIS */
            jistoeuc(ifd, ofd);
            break;

    case SJIS_CODE: /* SJIS */
            sjistoeuc(ifd, ofd);
            break;

    case EUC_CODE:  /* EUC */
    default:
            thru(ifd, ofd);
            break;
    }

    return 0;
}

/*
 */
int ToSJIS(int icode, int ifd, int ofd)
{
    switch(icode) {
    case JIS_CODE:  /* JIS */
            jistosjis(ifd, ofd);
            break;

    case EUC_CODE:  /* EUC */
            euctosjis(ifd, ofd);
            break;

    case SJIS_CODE: /* SJIS */
    default:
            thru(ifd, ofd);
            break;
    }

    return 0;
}

/*
 */
int ToJIS(int icode, int ifd, int ofd)
{
    switch(icode) {
    case SJIS_CODE: /* SJIS */
            sjistojis(ifd, ofd);
            break;

    case EUC_CODE:  /* EUC */
            euctojis(ifd, ofd);
            break;

    case JIS_CODE:  /* JIS */
    default:
            thru(ifd, ofd);
            break;
    }

    return 0;
}




#ifdef JFILTER_TEST
main(int argc, char **argv)
{
    int fd;
    int incode, outcode;
    char filename[1024];
    int i;
    int opt;
    int checkflag;
    int verbose;
    int filename_length;

    extern char *optarg;
    extern int  optind;

#if 0
    i = 1;
    if(argc>1) {
        if(isdigit(argv[1][0])) {
            outcode = atoi(argv[1]);
            i = 2;
        }
        else {
            outcode = -1;
        }
    }
#endif

    outcode = -1;
    checkflag = 0;
    verbose = 0;

    while((opt = getopt(argc, argv, "hcvjes")) != -1) {
        switch(opt) {
        case 'h':
            break;
        case 'c':
            checkflag = 1;
            break;
        case 'v':
            verbose = 1;
            break;
        case 'j':
            outcode = JIS_CODE;
            break;
        case 'e':
            outcode = EUC_CODE;
            break;
        case 's':
            outcode = SJIS_CODE;
            break;
        }
    }

    if(verbose && !checkflag) {
        fprintf(stderr, "convert to %s\n",
            outcode>=0 ? (outcode==JIS_CODE ? "JIS" :
                (outcode==EUC_CODE ? "EUC" :
                    (outcode==SJIS_CODE ? "SJIS" : "OTHER"))) : "UNKOWN" );
    }

    filename_length = 0;
    for(i=optind;i<argc;i++) {
        if(strlen(argv[i])>filename_length)
            filename_length = strlen(argv[i]);
    }
    for(i=optind;i<argc;i++) {
#if 0
        fprintf(stderr, "file %s ", argv[i]);
#endif
        incode = CheckFileKanjiCode(argv[i], 100);

        if(verbose && !checkflag) {
            fprintf(stderr, "%-*s %s\n", filename_length, argv[i],
                incode>=0 ? (incode==JIS_CODE ? "JIS" :
                    (incode==EUC_CODE ? "EUC" :
                        (incode==SJIS_CODE ? "SJIS" : "OTHER"))) : "UNKOWN" );
        }
        if(checkflag) {
            fprintf(stdout, "%-*s %s\n", filename_length, argv[i],
                incode>=0 ? (incode==JIS_CODE ? "JIS" :
                    (incode==EUC_CODE ? "EUC" :
                        (incode==SJIS_CODE ? "SJIS" : "OTHER"))) : "UNKOWN" );
            fflush(stdout);

            continue;
        }

        if(incode==-1) {
            fprintf(stderr, "*ignore* code\n");
            continue;
        }

        if((fd=open(argv[i],O_RDONLY))<0) {
            fprintf(stderr, "Cannot open %s\n.", argv[1]);
        }
        fflush(stderr);

        /* Convet from fd to STDOUT */
        switch(outcode) {
            case JIS_CODE:  ToJIS(incode, fd, 1);   break;
            case EUC_CODE:  ToEUC(incode, fd, 1);   break;
            case SJIS_CODE: ToSJIS(incode, fd, 1);  break;
            default:        thru(fd, 1);        break;
        }

        close(fd);
    }
    close(1);
}
#endif

#ifdef TEST
main(int argc, char **argv)
{
    int i, code;

    i = 1;
    while(i<argc) {
        code = CheckFileKanjiCode(argv[i], 66);
        printf("%s: %s\n", argv[i],
            code>=0 ? (code==0 ? "JIS" :
            (code==1 ? "SJIS" : (code==2 ? "EUC" : "OTHER"))) : "UNKOWN" );
        i++;
    }

    exit(WCOL_EXIT_NORMAL);
}
#endif

