/*
 *  WCOL: Message and Logging procedures
 *      by k-chinen@is.aist-nara.ac.jp, 1994, 1995
 *
 *  $Id: msg.c,v 1.11 1996/11/24 14:48:42 k-chinen Exp k-chinen $
 */

/*
 * Simple:
 *      Trace:  show message for debug.
 *      Log:    write message to log-file. (append newline)
 *
 * Complex:
 *      Error:  message. (append newline)
 *      Fatal:  message and log. (append newline)
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include <syslog.h>
#include <sys/time.h>
#include <sys/types.h>
#include <fcntl.h>
#include <time.h>
#ifdef TIME_WITH_SYS_TIME
#include <sys/time.h>
#endif
#include <errno.h>

#include "msg.h"


#define XPID    getpid()

/*
 * globals
 */
/* public */
int trace_flag = 0;
char log_file[LOG_FILENAMESIZE];

/* private */
static char log_name[LOG_FILENAMESIZE];
static char host_name[LOG_FILENAMESIZE];
static char *mon_str[]={"Jan","Feb","Mar","Apr","May","Jun",
                            "Jul","Aug","Sep","Oct","Nov","Dec"};



#ifdef __STDC__
int InitLogger(char *name)
#else
int InitLogger(name)
char *name;
#endif
{
    register char *p;

    strcpy(log_name, name);
#ifdef USE_SYSLOG
#ifdef USE_SYSLOG42
    /* for BSD 4.2 syslog (e.g., ULTRIX) */
    openlog(log_name, LOG_PID);
#else
    openlog(log_name, LOG_PID, LOG_FACILITY);
#endif
#endif

    if(gethostname(host_name, LOG_FILENAMESIZE)) {
        strcpy(host_name, "-");
    }
    else {
        /* get short name */
        p = host_name;
        while(*p && *p!='.') {
            p++;
        }
        if(*p=='.')
            *p = '\0';
    }

    return 0;
}


/*
 * Trace - show message if 'trace_flag' is up.
 */
/*
 * RawTrace - show message and flush it.
 *          main body of Trace() (see msg.h).
 */

#ifdef __STDC__
int RawTrace(const char *fmt, ...)
#else
int RawTrace(va_alist)
va_dcl
#endif
{
#ifndef __STDC__
    char *fmt;
#endif
    va_list args;

#ifdef __STDC__
    va_start(args,fmt);
#else
    va_start(args);
    fmt = va_arg(args, char *);
#endif

    vfprintf(stderr,fmt,args);
    fflush(stderr);

    va_end(args);

    return 0;
}


/*
 * Error - show error messages like as Trace but it is not check 'trace_flag'.
 *         Because error must be shown.
 */

#ifdef __STDC__
int Error(const char *fmt, ...)
#else
int Error(va_alist)
va_dcl
#endif
{
#ifndef __STDC__
    char *fmt;
#endif
    va_list args;
    char report[LOG_LINESIZE];

#ifdef __STDC__
    va_start(args,fmt);
#else
    va_start(args);
    fmt = va_arg(args, char *);
#endif

    vsprintf(report, fmt, args);

    fprintf(stderr, "### ERROR %d> %s\n", XPID, report);
    fflush(stderr);

    va_end(args);

    return 0;
}


/*
 * RawLog --- logging procedure; write line to log-file or send to syslog
 */
static int RawLog(int level, char *rline)
{
#ifdef USE_SYSLOG
    syslog(level, "%s", rline);
#else

    int log_fd;
    time_t now;
    struct tm *ntm;
    char line[LOG_LINESIZE];

    if(rline[0]=='\0') {
        fprintf(stderr, "RawLog: report is empty\n");
        return -1;
    }

    time(&now);
#ifdef LOGGING_LOCALTIME
    ntm = localtime(&now);      /* get time with local time */
#else
    ntm = gmtime(&now);         /* get time with UTC (a.k.a. GMT) */
#endif

    /* open log file. if it is fail, try to open console */
    if((log_fd=open(log_file, O_WRONLY|O_CREAT|O_APPEND, 0644)) < 0 ) {
        if((log_fd=open("/dev/console", O_WRONLY|O_APPEND))<0) {
            fprintf(stderr, "Sorry, cannot open logfile and console\n");
            return -1;
        }
        else {
            sprintf(line, "wcol: fail to open logfile-'%s'.\n", log_file);
            write(log_fd, line, strlen(line));
        }
    }

    /* month day time host program[pid]: message */
    sprintf(line, "%s %02d %02d:%02d:%02d %s %s[%d]: %s\n",
        mon_str[ntm->tm_mon], ntm->tm_mday,
        ntm->tm_hour, ntm->tm_min, ntm->tm_sec,
        host_name, log_name, XPID, rline);
    write(log_fd, line, strlen(line));
    close(log_fd);

#endif /* USE_SYSLOG */

    return 0;
}


/*
 * Log ---
 */
#ifdef __STDC__
int Log(const char *fmt, ...)
#else
int Log(va_alist)
va_dcl
#endif
{
    va_list args;
#ifndef __STDC__
    char *fmt;
#endif
    char report[LOG_LINESIZE];

#ifdef __STDC__
    va_start(args,fmt);
#else
    va_start(args);
    fmt = va_arg(args, char *);
#endif

    vsprintf(report, fmt, args);

    if(report[0]=='\0') {
        fprintf(stderr, "Log: report is empty\n");
    }
    else {
        RawLog(LOG_NOTICE, report);
    }

    if(trace_flag) {
        fprintf(stderr,"=== LOG %d> %s\n", XPID, report);
        fflush(stderr);
    }

    va_end(args);

    return 0;
}

/*
 * DLog ---
 */
#ifdef __STDC__
int DLog(const char *fmt, ...)
#else
int DLog(va_alist)
va_dcl
#endif
{
    va_list args;
#ifndef __STDC__
    char *fmt;
#endif
    char report[LOG_LINESIZE];
    struct timeval dnow;
    char *p;

#ifdef __STDC__
    va_start(args,fmt);
#else
    va_start(args);
    fmt = va_arg(args, char *);
#endif

    gettimeofday(&dnow, NULL);
    sprintf(report, "%d.%06d ", dnow.tv_sec, dnow.tv_usec);
    p = report;
    while(*p) {
        p++;
    }

    vsprintf(p, fmt, args);

    if(report[0]=='\0') {
        fprintf(stderr, "Log: report is empty\n");
    }
    else {
        RawLog(LOG_NOTICE, report);
    }

    if(trace_flag) {
        fprintf(stderr,"=== LOG %d> %s\n", XPID, report);
        fflush(stderr);
    }

    va_end(args);

    return 0;
}



/*
 * Fatal ---
 */
#ifdef __STDC__
int Fatal(const char *fmt, ...)
#else
int Fatal(va_alist)
va_dcl
#endif
{
    va_list args;
#ifndef __STDC__
    char *fmt;
#endif
    char report[LOG_LINESIZE];
    char *p;

#ifdef __STDC__
    va_start(args,fmt);
#else
    va_start(args);
    fmt = va_arg(args, char *);
#endif

    /* append prefix */
    strcpy(report, "FATAL ");
    p = report + strlen(report);
    vsprintf(p, fmt, args);

    RawLog(LOG_WARNING, report);

    fprintf(stderr,"@@@ FATAL %d> %s\n", XPID, report);
    fflush(stderr);

    va_end(args);

    return 0;
}


int SwitchLog(char *logpath, char *note)
{
    char logfname[LOG_FILENAMESIZE];
    int logfd;
    time_t ltime;
    time_t utime;


    /* get local and UTC date/time */
    time(&ltime);
    utime = mktime(gmtime(&ltime));

/*
    wcol_pid = getpid();
*/

    sprintf(logfname, "%s/log-%d", logpath, XPID);
    if((logfd = open(logfname, O_WRONLY|O_CREAT, 0644))<0) {
        Trace("SwitchLog: Cannot open new logfile-'%s'\n", logfname);
        return 1;
    }
    if(dup2(logfd, 2)<0) {
        Trace("SwitchLog: fail dup2() to stderr\n");
        return 1;
    }
#if 1
    if(dup2(logfd, 1)<0) {
        Trace("SwitchLog: fail dup2() to stdout\n");
        return 1;
    }
#endif


    /*
     * print information in head
     */
    if(note==NULL || *note == '\0') {
        Trace(";\n; log for wcol module\n;\n");
    }
    else {
        Trace(";\n; %s\n;\n", note);
    }
    Trace("; Pid %d, Ppid %d\n", getpid(), getppid());
    Trace("; Local %d : %s", ltime, ctime(&ltime));
    Trace("; UTC   %d : %s", utime, ctime(&utime));
    Trace(";\n");

    return 0;
}




