/*
 * wcolE, session.c - Session management procedures
 *          by k-chinen@is.aist-nara.ac.jp, 1995, 1996, 1997
 *
 * NOTE: session.c is appeared wcolC. This is new version of session.c
 *
 * $Id: session.c,v 1.13 1997/09/15 01:57:57 k-chinen Exp k-chinen $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


/*
 * phase translation:
 *
 *  phase name      function
 *  -----------------------------------------------------------------
 *
 *  PHASE_IDLE
 *      |           StartSession
 *  PHASE_START
 *  PHASE_REQUEST                           +
 *      |           ConnectSiblingSession   |
 *  PHASE_CONNECTSIBLING                    | ConnectQuerySiblingSession
 *      |           QuerySession            |
 *  PHASE_QUERY                             +
 *      |           ConnectSession
 *  PHASE_CONNECTSERVER
 *      |           ReachSession
 *  PHASE_HEADER    ReadSession
 *  PHASE_BODY
 *      |           CloseSession
 *  PHASE_END
 *      |
 *  PHASE_IDLE
 *
 */



#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/uio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <limits.h>
#include <time.h>
#ifdef TIME_WITH_SYS_TIME
#include <sys/time.h>
#endif
#include <signal.h>
#include <setjmp.h>
#include <sys/stat.h>
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <errno.h>



#include "wcol.h"
#include "sigs.h"
#include "fds.h"
#include "url.h"
#include "hname.h"
#include "socket.h"

#include "request.h"
#include "http.h"
#include "send.h"
#include "parse.h"
#include "session.h"
#include "handler.h"
#include "manager.h"
#include "sibling.h"

#include "ftphand.h"

#include "misc.h"


#define DEBUG_SESSION
#define _TRACE_POS
#define _TRACE_BUF



#ifndef WCOL_FD_MAX
#define WCOL_FD_MAX     FD_SETSIZE
#endif /* WCOL_FD_MAX */


#ifndef REPORT_TIMEOUT
#define REPORT_TIMEOUT              (1)     /* 1 sec */
#endif




#define SESSION_SUB_TIME(s, x, y) \
        ( (s->x.tv_sec  - s->y.tv_sec )*1000 \
        + (s->x.tv_usec - s->y.tv_usec)/1000 )


#define ServerTypeStr(v) ((v>=SERVER_UNKNOW&&v<=SERVER_PARENT) ? \
            _server_type_str[v] : _server_type_str[SERVER_UNKNOW])


char *_server_type_str[] = {"UNKNOW", "ORIGIN", "SIBLING", "PARENT"};



extern int catalyst_flag;
extern url_t *fetch_parent, *prefetch_parent, *import_sibling;

extern Session *session_pool;
extern int handler_no;
extern modinfo_t *handlers;

extern int reception_timeout;


#if 0
int     request_seqno = 0;
int     session_seqno = 0;
#endif
int     *request_seqno;
int     *session_seqno;


Session *session_list;
int     max_session = MAX_SESSION;
int     num_session = 0;
int     num_client_session   = 0;
int     num_prefetch_session = 0;
int     num_import_session   = 0;


fdset_t session_fdset;
fd2sess_t fd2sess[WCOL_FD_MAX];
int     num_fd  = 0;    /* number of connection(s), file descriptor(s) */
int     last_fd = 0;    /* last connections, file descriptor */

static int  icmp_fd = -1;

struct timeval session_currenttime;

int session_timeout, session_lifetime;    /* msec */
int query_timeout;      /* msec */




static int most_last_fd     = -1;
static int most_num_fd      = -1;
static int most_num_session = -1;


int LogStatus()
{
    int i;
    char rt=' ';
    static int already_report=0;

    if(most_last_fd<last_fd) {
        most_last_fd = last_fd;
        rt = 'l';
    }
    if(most_num_fd<num_fd) {
        most_num_fd = num_fd;
        rt = 'n';
    }
    if(most_num_session<num_session) {
        most_num_session = num_session;
        rt = 's';
    }

    if(!already_report && *session_seqno % 100 == 0) {
        rt = '*';
    }

    if(rt!=' ') {
#if 0
      Log(";            sess#   req# : T :  session  :  last_fd  :  num_fd");
/*                     123456 123456 : 1 : 1234 1234 : 1234 1234 : 1234 1234 */
#endif
      DLog("INFO FACTOR %6d %6d : %c : %4d %4d : %4d %4d : %4d %4d",
            *session_seqno, *request_seqno,
            rt,
            num_session, most_num_session,
            last_fd, most_last_fd, num_fd, most_num_fd);
    }

    already_report=1;

    return 1;
}




/*
 *  return
 *      0   not found
 *      !0  found
 */

int IsInSesssionPool(char *xurl, Session *self)
{
    register int i;
    register Session *sp;

    if(xurl==NULL || *xurl=='\0') {
        return 0;
    }

    for(i=0;i<MAX_SESSION*MAX_HANDLER;i++) {
        sp = &session_pool[i];
        if(sp!=self) {
            if(sp->phase!=PHASE_IDLE && sp->_url[0] && *xurl) {
                if(strcmp(sp->_url, xurl)==0) {
                    Log("; %s of %d is found in %d", xurl, handler_no, i);
                    return i;
                }
            }
        }
    }

    return 0;
}



/*
 * InitSessions
 */
void InitSessions(int http_fd, int command_fd)
{
    int i;

    session_list = &session_pool[MAX_SESSION*handler_no+0];
    for(i=0;i<MAX_SESSION;i++) {
        session_list[i].type  = SESS_NOP;
        session_list[i].phase = PHASE_IDLE;
    }

    last_fd = 0;
    for(i=0;i<FD_SETSIZE;i++) {
        fd2sess[i].type             = CONN_NOP;
        fd2sess[i].timeout.tv_sec   = 0;
        fd2sess[i].timeout.tv_usec  = 0;
        fd2sess[i].sess             = NULL;
        fd2sess[i].sibling_id       = -1;
    }

    /* */
    i = http_fd;
    fd2sess[i].type             = CONN_HTTPLISTEN;
    fd2sess[i].timeout.tv_sec   = 0;
    fd2sess[i].timeout.tv_usec  = 0;
    fd2sess[i].sess             = NULL;
    fd2sess[i].sibling_id       = -1;

    /* */
    i = command_fd;
    fd2sess[i].type             = CONN_COMMAND;
    fd2sess[i].timeout.tv_sec   = 0;
    fd2sess[i].timeout.tv_usec  = 0;
    fd2sess[i].sess             = NULL;
    fd2sess[i].sibling_id       = -1;


    /***
     ***
     ***/
    Trace("max session = %d (compiled %d)\n", max_session, MAX_SESSION);
    Trace("num of FDs  = %d/%d\n", WCOL_FD_MAX, FD_SETSIZE);
    Trace("HTTP fd#    = %4d\n", http_fd);
    Trace("command fd# = %4d\n", command_fd);

    FD_ZERO(&session_fdset);


#ifdef USE_ICMP
    if((icmp_fd = InitICMP())>=0) {
        FD_SET(icmp_fd, &session_fdset);
    }
#endif
}


int CheckLastFd()
{
    int i;
    int c;

    num_fd = 0;
    c = -1;
    for(i=0;i<FD_SETSIZE;i++) {
        if(fd2sess[i].type != CONN_NOP) {
            c = i;
            if(fd2sess[i].type != CONN_HTTPLISTEN &&
                fd2sess[i].type != CONN_COMMAND) {
                num_fd++;
            }
        }
    }
    last_fd = c;

    return last_fd;
}




/*
 * RawDisplaySessions - dispaly sessions
 *      all - print all sessions or not idle sessions only
 */
static void RawDisplaySessions(Session session_list[], int all)
{
    Session *s;
    int i;
    int c;

#if 1
    /*
     * This routine only display procedures. So, if 'trace_flag' is not
     * set, we don't need these procedures.
     */
    if(!trace_flag) {
        return;
    }

    fflush(stdout);
    fflush(stderr);
#endif

    /* count sessions */
    c = 0;
    for(i=0;i<MAX_SESSION;i++) {
        if(session_list[i].phase != PHASE_IDLE) {
            c++;
        }
    }

#if 0
    handlers[handler_no].busy = c;
#endif

    if(c) {
        Trace("* Sessions %d (handler# %d) %10d.%06d, life %ds, idle %ds\n",
            wcol_pid, handler_no,
            session_currenttime.tv_sec, session_currenttime.tv_usec,
            session_lifetime/1000,
            session_timeout/1000
        );
        Trace("* count %d, session %d [c %d, p %d, i %d], file %d\n",
            c, num_session, num_client_session,
            num_prefetch_session, num_import_session, last_fd);




#if 1
        /***
         *** FD
         ***/
        /*     12345   12345678 1 1 1234567 1234567 123 123 123 123 1234567 1234.. */
        Trace("sess#::   addr   T P   life    idle   c , s , h , b    len   URL\n");
        for(i=0;i<MAX_SESSION;i++) {
            s = &session_list[i];
            if(all || s->phase!=PHASE_IDLE) {
                Trace("%5d:: %8x %c %c %7d %7d %3d,%3d,%3d,%3d %7ld %s\n",
                    i, s, ReqModeChar(s->type), SessPhaseChar(s->phase),
                WCOL_SUB_TIME(session_currenttime, s->begin),
                WCOL_SUB_TIME(session_currenttime, s->hot),
                    s->client_fd, s->server_fd, s->header_fd, s->body_fd,
                    (long)(s->hsum+s->bsum),
                    s->_url);

#if 0
            Trace("_URL: x %3d %p '%s'\n", i, s, s->_url);
#endif
            }

        }

#endif

#if 0
        /***
         *** FD
         ***/
        /*     12345   12345678 1 1 12345 123 123 123 123 1234567 1234.. */
        Trace("sess#::   addr   T P  sid   c , s , h , b    len   URL\n");
        for(i=0;i<MAX_SESSION;i++) {
            s = &session_list[i];
            if(all || s->phase!=PHASE_IDLE) {
                Trace("%5d:: %8x %c %c %5d %3d,%3d,%3d,%3d %7d %s\n",
                    i, s, ReqModeChar(s->type), SessPhaseChar(s->phase),
                    s->req.sid,
                    s->client_fd, s->server_fd, s->header_fd, s->body_fd,
                    s->hsum+s->bsum,
                    s->_url);
            }
        }
#endif

#if 0
        /***
         *** type, phase and time
         ***/
        /*     12345   1 1 123456 1234567890 1234567 1 1234 */
        Trace("sess#:: T P   sid   lap-time    len   h URL\n");

        for(i=0;i<MAX_SESSION;i++) {
            s = &session_list[i];

            if(all || s->phase==PHASE_IDLE) {
#if 0
            Trace(  "%5d:: %c %c %6s %10s %7d * %s\n",
                i, ReqModeChar(s->type), SessPhaseChar(s->phase),
                "-",
                "-",
                s->hsum+s->bsum,
                s->_url);
#endif
                continue;
            }

            if(s->url != NULL) {
            Trace("%5d:: %c %c %6d %10d %7d - %s\n",
                i, ReqModeChar(s->type), SessPhaseChar(s->phase),
                s->req.sid,
                WCOL_SUB_TIME(session_currenttime, s->hot),
                s->hsum+s->bsum,
                s->url->full_loc);
            }
            else {
            Trace("%5d:: %c %c %6d %10d %7d * %s\n",
                i, ReqModeChar(s->type), SessPhaseChar(s->phase),
                s->req.sid,
                WCOL_SUB_TIME(session_currenttime, s->hot),
                s->hsum+s->bsum,
                s->_url);
            }
        }
#endif


    }
    else {
        Trace("* Sessions %d %10d.%06d %d : no session\n",
            wcol_pid, session_currenttime.tv_sec, session_currenttime.tv_usec,
            session_timeout);
    }




    /* count fd */
    c = 0;
    for(i=0;i<FD_SETSIZE;i++) {
        if(fd2sess[i].type<=CONN_NOP) {
            /* nothing */
        }
        else {
            c++;
        }
    }

    if(c) {
        Trace("\n");
 Trace("fd# :@   addr    T   ID timeout(sec.usec) / last_fd %d / num_fd %d\n",
            last_fd, num_fd);
#if 0
        for(i=0;i<last_fd;i++) {
            if(fd2sess[i].type>CONN_NOP) {
                Trace("%4d:@ %8x %3d %3d %10d.%06d\n", i,
                    fd2sess[i].sess, fd2sess[i].type, fd2sess[i].sibling_id,
                    fd2sess[i].timeout.tv_sec, fd2sess[i].timeout.tv_usec
                    );
            }
        }
#else
        for(i=0;i<FD_SETSIZE;i++) {
            if(fd2sess[i].type<=CONN_NOP) {
#if 0
                Trace("%4d:# #p %3d %3d %10d.%06d\n", i,
                    fd2sess[i].sess, fd2sess[i].type, fd2sess[i].sibling_id,
                    fd2sess[i].timeout.tv_sec, fd2sess[i].timeout.tv_usec
                    );
#endif
            }
            else {
                Trace("%4d:@ %8x %3d %3d %10d.%06d\n", i,
                    fd2sess[i].sess, fd2sess[i].type, fd2sess[i].sibling_id,
                    fd2sess[i].timeout.tv_sec, fd2sess[i].timeout.tv_usec
                    );
            }
        }
#endif

    }
    else {
        Trace("no handle fd / %d\n", last_fd);
    }

    Trace("-------------------------------\n");


#if 1
    fflush(stdout);
    fflush(stderr);
#endif

    return;
}


void DisplaySessions(int all)
{
#if 1
    RawDisplaySessions(session_list, all);
#else
    int i;
    Session *ls[MAX_SESSION];

    for(i=0;i<MAX_SESSION;i++) {
        ls[i] = &session_list[i];
    }

    RawDisplaySessions(ls, all);
#endif
}






static int _session_counter_ = 0;


int CleanSession(Session *sess)
{
    sess->type      = SESS_NOP;
    sess->phase     = PHASE_IDLE;

    sess->server_type = SERVER_UNKNOW;
    sess->server_name[0] = '\0';
    sess->server_port = -1;

    sess->server_fd = -1;
    sess->client_fd = -1;
    sess->header_fd = -1;
    sess->body_fd   = -1;

    sess->len       = 0;
    sess->clen      = -1;
    sess->wlen      = 0;
    sess->hsum      = 0;
    sess->bsum      = 0;

    memset(sess->request_buf,sizeof(sess->request_buf), 0);
    memset(sess->header_buf, sizeof(sess->header_buf),  0);
    memset(sess->body_buf,   sizeof(sess->body_buf),    0);
    memset(sess->linebuf,    sizeof(sess->linebuf),     0);

    sess->begin.tv_sec      = 0;
    sess->begin.tv_usec     = 0;
    sess->hot.tv_sec        = 0;
    sess->hot.tv_usec       = 0;

    sess->use_full_location = 0;
    sess->url       = NULL;
    sess->_url[0]   = '\0';

    ClearInfo(&(sess->info));

    sess->rcv_mode  = -1;

    sess->query_sibling = 0;

    return 0;
}


/*
 * NewSession - alloc structure-'Session' to heap
 */
Session *NewSession()
{
    int     i;

    for(i=0;i<MAX_SESSION;i++) {
        if(session_list[i].type==SESS_NOP) {

            CleanSession(&session_list[i]);

            return &session_list[i];
        }
    }

    Log("; no idle session");

    return NULL;
}

Session *AltanativeSession()
{
    int     i;

    for(i=0;i<MAX_SESSION;i++) {
        if(session_list[i].type==SESS_PREFETCH) {

#if 0
            CleanSession(&session_list[i]);
#endif

            return &session_list[i];
        }
    }

    Log("; no prefetch session");

    return NULL;
}


/*
 * DeleteSession - Free memory and close files.
 */
void DeleteSession(Session *sess)
{
    if(sess==NULL)
        return;

#if 1
    Trace("DeleteSession: session %#x, try close [%d,%d,%d,%d]\n", sess,
        sess->client_fd, sess->server_fd, sess->header_fd, sess->body_fd);
#endif

    release_socket(&(sess->client_fd));
    release_socket(&(sess->server_fd));
    if(sess->header_fd!=-1) {
        close(sess->header_fd);
        sess->header_fd = -1;
    }
    if(sess->body_fd!=-1) {
        close(sess->body_fd);
        sess->body_fd = -1;
    }

#if 0
    Trace("DeleteSession: session %#x, close [%d,%d,%d,%d]\n", sess,
        sess->client_fd, sess->server_fd, sess->header_fd, sess->body_fd);
#endif

    URL_Delete(sess->url);
    sess->url = NULL;
    sess->_url[0] = '\0';

    sess->type  = SESS_NOP;
    sess->phase = PHASE_IDLE;

#ifdef MEMORY_CHECK
    show_memoryusage();
#endif


    return;
}




/*
 *
 */
int JoinSessionList(Session *sess)
{
    int i;
    int t_us, t_s;


    Trace("JoinSessionList: sess %p\n", sess);
    if(sess==NULL) {
        return 0;
    }

    if(sess->server_fd<0) {
        Error("JoinSessionList: sess %p is not connected\n", sess);
        return 0;
    }

    /* Append fd to session_fdset */
    FD_SET(sess->server_fd, &session_fdset);

    fd2sess[sess->server_fd].type       = CONN_HTTP;
    fd2sess[sess->server_fd].sess       = sess;
    fd2sess[sess->server_fd].sibling_id = -1;

    fd2sess[sess->server_fd].timeout = session_currenttime;
    t_us = fd2sess[sess->server_fd].timeout.tv_usec;
    t_us += (session_timeout%1000)*1000;
    t_s = (t_us > 1000000);
    fd2sess[sess->server_fd].timeout.tv_usec = t_us % 1000000;

    fd2sess[sess->server_fd].timeout.tv_sec +=
        t_s + session_timeout/1000;


#ifdef DEBUG_SESSION
    if(sess->url!=NULL) {
    i = -1234;
    Trace("JoinSessionList:  session# %4d, [%4d,%4d,%4d,%4d] '%s'\n",
            i, sess->client_fd, sess->server_fd, sess->header_fd, sess->body_fd
            , sess->url->full_loc);
    }
    else {
    Trace("JoinSessionList:  session# %4d, [%4d,%4d,%4d,%4d] '%s'\n",
            i, sess->client_fd, sess->server_fd, sess->header_fd, sess->body_fd
            , sess->_url);
    }
#endif

#if 1
    DisplaySessions(0);
#endif

    CheckLastFd();

    return 0;
}


/*
 *
 */
void LeaveSessionList(Session *sess)
{

    Trace("LeaveSessionList: sess %p\n", sess);

    if(sess->server_fd >= 0) {
        FD_CLR(sess->server_fd, &session_fdset);
        fd2sess[sess->server_fd].type       = CONN_NOP;
        fd2sess[sess->server_fd].sess       = NULL;
        fd2sess[sess->server_fd].sibling_id = -1;

#ifdef DEBUG_SESSION
        Trace("LeaveSessionList: session# %4d, [%4d,%4d,%4d,%4d]\n", -1,
            sess->client_fd, sess->server_fd, sess->header_fd, sess->body_fd);
#endif

        CheckLastFd();

#if 1
        DisplaySessions(0);
#endif
    }

    return;
}


/*
 *
 */
int LeaveWaitSessionList(Session *sess)
{
    /*
        LeaveSessionList(sess);
    */
    return 0;
}







/*
 * ErrorAbortSession - send error message and abort
 *      responce_type - responce to commander
 *      responce_cod - responce to client
 *      short_msg - short message
 *      long_msg - long message
 */
int ErrorAbortSession(Session *sess,
        int responce_type,
        int responce_code, char *short_msg, char *long_msg)
{
    hreport_t report;
    off_t wlen;
    struct timeval error_end;
    char *label;

    if(sess==NULL) {
        Error("ErrorAbortSession: null pointer");
        return -1;
    }

    sess->phase = PHASE_ERROR;

    gettimeofday(&(sess->end_tp), NULL);

#if 0
    if(sess->url==NULL) {
    Trace("ErrorAbortSession: code %3d, null-URL\n\treport %d, sibling %d\n",
        responce_code,
        responce_type, sess->query_sibling
        );
    }
    else {
    Trace("ErrorAbortSession: code %3d, URL '%s'\n\treport %d, sibling %d\n",
        responce_code, sess->url->full_loc,
        responce_type, sess->query_sibling
        );
    }
#endif
    Trace("ErrorAbortSession: code %3d, URL '%s'\n\treport %d, sibling %d\n",
        responce_code, sess->_url,
        responce_type, sess->query_sibling
        );

    if(sess->server_fd>=0) {
        LeaveSessionList(sess);
        release_socket(&(sess->server_fd));
    }

    wlen = (off_t)0;
    if(sess->client_fd>=0) {
        wlen = HTTP_ErrorMessage_Old(sess->client_fd,
                responce_code, short_msg, long_msg, sess->req.ver_major>0);
        release_socket(&(sess->client_fd));

        if(wlen <(off_t) 0) {
            wlen = 0;
        }
    }

    if(sess->header_fd>=0) {
        ftruncate(sess->header_fd, sess->hsum);
        close(sess->header_fd);
        sess->header_fd = -1;
    }
    if(sess->body_fd>=0) {
        ftruncate(sess->body_fd, sess->bsum);
        close(sess->body_fd);
        sess->body_fd = -1;
    }

    if(sess->query_sibling) {
    }

    gettimeofday(&error_end, NULL);

    label="ERROR-UNKNOW";
    if(sess->type==SESS_PREFETCH) {
        Log("; error of self prefetch session");
        label = "ERROR-PREFETCH";
    }
    else
    if(sess->type==SESS_CLIENT) {
        Log("; error of client's fetch session");
        label = "ERROR-CLIENT";
    }
    else
    if(sess->type==SESS_IMPORT) {
        Log("; error of client's fetch session");
        label = "ERROR-IMPORT";
    }
    else {
        Log("; error of unknown session [type %d]", sess->type);
        label="ERROR-UNKNOW";
    }
    Log("%d.%06d %s %d %s HTTP %d.%d %s %s %s %s %s %d %s %ld %d :",
        error_end.tv_sec, error_end.tv_usec, label,
        sess->req.hereport, sess->req.peername,
        sess->req.ver_major, sess->req.ver_minor,
        sess->req.nocache == 1 ? "n" : "-",
        sess->req.ims == -1 ? "-" : "*",
        sess->req.auth[0] == '\0' ? "-" : "X",
        ReqMethodStr(sess->req.method), sess->req.url,
        responce_code,
        "-",
        (long)wlen,
        WCOL_SUB_TIME(error_end, sess->req.issue_time)
    );


    Log("; aborted by error  (%d) %d %s %d %ld %ld <%s>",
            responce_type, responce_code,
            /*
            sess->url->full_loc,
            */
            sess->_url,
            WCOL_SUB_TIME(sess->end_tp, sess->start_tp),
            (long)sess->hsum, (long)sess->bsum,
            short_msg);


    Trace("ErrorAbortSession: END #%d, len %d, type %d\n",
        sess->server_fd, sess->hsum+sess->bsum, sess->type);

#if 1
    memset(&report, 0, sizeof(report));
#endif

    report.type             = responce_type;
    report.sid              = sess->req.sid;
    report.id               = sess->req.id;
    report.keep_alive       = 0;
    ShowHReport(&report);


    /*
     * To avoid deadlock with IPC, set timer
     */
    pipe_broken = 0;
    XSendRecord(REPORT_TIMEOUT, sess->command_fd, &report, sizeof(hreport_t));


    DeleteSession(sess);

    return 0;
}



/*
 * Connect origin server or other proxy server.
 */
int ConnectSession(Session *sess)
{
    url_t *url;
    char  *request_url;
    char  *request_host;
    int   request_port;
    hreport_t report;


    sess->hot       = session_currenttime;
    sess->phase     = PHASE_CONNECTSERVER;

#if 0
    ShowRequest(sess->req);
#endif
    URL_Normalize(sess->url);
    URL_Show(sess->url);
    url = sess->url;

    if(catalyst_flag && sess->req.mode == REQ_IMPORT) {
        if(import_sibling) {
            request_host    = import_sibling->host;
            request_port    = atoi(import_sibling->port);
            request_url     = url->full_loc;
            sess->use_full_location = 1;

            sess->server_type = SERVER_SIBLING;
        }
        else
        if(fetch_parent) {
            request_host    = fetch_parent->host;
            request_port    = atoi(fetch_parent->port);
            request_url     = url->full_loc;
            sess->use_full_location = 1;

            sess->server_type = SERVER_PARENT;
        }
        else {
            request_host    = url->host;
            request_port    = atoi(url->port);
            request_url     = url->path;
            sess->use_full_location = 0;

            sess->server_type = SERVER_ORIGIN;
        }
    }
#if 0
    else
    if(sess->req.mode == REQ_PREFETCH) {
        if(prefetch_parent) {
            request_host    = prefetch_parent->host;
            request_port    = atoi(prefetch_parent->port);
            request_url     = url->full_loc;
            sess->use_full_location = 1;

            sess->server_type = SERVER_PARENT;
        }
        else {
            request_host    = url->host;
            request_port    = atoi(url->port);
            request_url     = url->path;
            sess->use_full_location = 0;

            sess->server_type = SERVER_ORIGIN;
        }
    }
#endif
    else {
        if(fetch_parent) {
            request_host    = fetch_parent->host;
            request_port    = atoi(fetch_parent->port);
            request_url     = url->full_loc;
            sess->use_full_location = 1;

            sess->server_type = SERVER_PARENT;
        }
        else {
            request_host    = url->host;
            request_port    = atoi(url->port);
            request_url     = url->path;
            sess->use_full_location = 0;

            sess->server_type = SERVER_ORIGIN;
        }
    }

    strcpy(sess->server_name, request_host);
    sess->server_port = request_port;

#if 0
    DisplaySessions(0);
#endif


    if(SetupFile(sess->body_fn, sess->url->full_loc, BODY_TMP_FILE,
            &(sess->body_fd))) {
        return 1;
    }

#if 1
    Trace("ConnectSession: sess %p, [%3d,%3d,%3d,%3d]\n",
        sess, sess->client_fd, sess->server_fd, sess->header_fd, sess->body_fd);
    Trace("\tServer='%s', Port=%d, Path='%s'\n",
        request_host, request_port, request_url);

    DisplaySessions(0);
#endif

    gettimeofday(&(sess->start_tp), NULL);


    sess->phase     = PHASE_NAMERESOLV;

    /*
     * connect server
     */
    errno = 0;
    if((sess->server_fd = TCP_aconnect(request_host, request_port))<0) {

        Error("ConnectSession: Cannot connect to '%s' (%d)",
            request_host, errno );

        if(sess->server_fd == WCOL_SOCK_SOCKETERROR) {
            ErrorAbortSession(sess, HREP_INNERERROR,
                500, "Socket Error", "Cannot make socket");
        }
        else
        if(sess->server_fd == WCOL_SOCK_UNKNOWHOST) {
            ErrorAbortSession(sess, HREP_IGNOREREQUEST,
                403, "Host Resolution Error",
                "Unknown host. please check it once more");
        }
        else
        if(sess->server_fd == WCOL_SOCK_CONNECTERROR) {
            ErrorAbortSession(sess, HREP_IGNOREREQUEST,
                403, "Connect Error",
                "Cannot connect server. Server is down or busy");
        }
        else
        {
            ErrorAbortSession(sess, HREP_IGNOREREQUEST,
                403, "Unknown Connect Error",
                "Sorry, I don't know what happend");
        }

        return 1;
    }
    sess->phase = PHASE_CONNECTSERVER;

#if 0
    Trace("ConnectSession: return 0\n");
#endif

    return 0;
}




int ConnectSession_Sibling(Session *sess, sibling_info_t *keeper)
{
    url_t *url;
    char  *request_url;
    char  *request_host;
    int   request_port;
    hreport_t report;


    if(keeper==NULL) {
        return ConnectSession(sess);
    }


    sess->hot       = session_currenttime;
    sess->phase     = PHASE_CONNECTSERVER;

#if 0
    ShowRequest(sess->req);
#endif
    URL_Normalize(sess->url);
    URL_Show(sess->url);
    url = sess->url;

    request_host    = keeper->name;
    request_port    = keeper->http_port;
    request_url     = url->full_loc;

    sess->use_full_location = 1;
    sess->server_type   = SERVER_SIBLING;
    strcpy(sess->server_name, keeper->name);
    sess->server_port   = keeper->http_port;


#if 1
    Trace("ConnectSession_Sibling:\n\tServer='%s', Port=%d, Path='%s'\n",
            request_host, request_port, request_url);
#endif


    if(SetupFile(sess->body_fn, sess->url->full_loc, BODY_TMP_FILE,
            &(sess->body_fd))) {
        return 1;
    }

    gettimeofday(&(sess->start_tp), NULL);

    /*
     * connect server
     */
    errno = 0;
    if((sess->server_fd = TCP_aconnect(request_host, request_port))<0) {
        Error("ConnectSession_Sibling: Cannot connent to '%s'", request_host);

        ErrorAbortSession(sess, HREP_IGNOREREQUEST,
            403, "Connect Error", "Cannot connect sibling");

        return 1;
    }
    sess->phase = PHASE_CONNECTSERVER;

    return 0;
}




int SendPostRequest(int sfd, char *path, char *orig_buf)
{
    char *cpos;
    char line[BUF_SIZE];
    char post_buf[BUF_SIZE];
    int  lno;
    int  nwrite;

#ifdef TRACE_SEND_REQUEST
    Trace("SendPostRequest: buffer length %d\n", strlen(orig_buf));
    dump_stringchunk("orig", orig_buf);
#endif

    post_buf[0] = '\0';

    lno = 0;
    cpos = orig_buf;
    while((cpos = sgets(orig_buf, cpos, line, BUF_SIZE))!=NULL) {
        if(lno==0) {
            sprintf(line, "POST %s HTTP/1.0\r\n", path);
        }
        strcat(post_buf, line);

        lno++;
    }

#ifdef TRACE_SEND_REQUEST
    Trace("SendPostRequest: buffer length %d\n", strlen(post_buf));
    dump_stringchunk("post", post_buf);
#endif

    nwrite = write(sfd, post_buf, strlen(post_buf));
    if(nwrite<0) {
        Error("SendPostRequest: write() fail (%d)", errno);
    }

    return nwrite;
}


/*
 *
 * return:
 *      success 0
 *      fail    !0
 */
int ReachSession(Session *sess)
{
    if(sess->phase!=PHASE_CONNECTSERVER) {
        Error("ReachSession: strang. %p is not PHASE_CONNECTSERVER", sess);
        return -1;
    }
    Trace("ReachSession: enter %d/%d, server %d\n",
            sess->req.sid, sess->req.id, sess->server_fd);

    /*
     * Record connection establish time.
     */
    gettimeofday(&(sess->conn_tp), NULL);

    if(sess->req.method==METHOD_GET) {

        pipe_broken = 0;
        if(sess->server_type==SERVER_ORIGIN) {
            if(sess->request_buf[0]) {
                HTTP_CookedRequest(sess->server_fd, sess->req.nocache,
                    sess->url->host, sess->url->path, sess->request_buf);
            }
            else {
                HTTP_NewRequest(sess->server_fd,
                    sess->url->host, sess->url->path,
                    sess->req.auth, sess->req.cookie,
					sess->req.nocache ,sess->req.noprefetch);
            }
        }
        else {
            if(sess->request_buf[0]) {
                HTTP_CookedRequest(sess->server_fd, sess->req.nocache,
                    sess->url->host, sess->url->full_loc, sess->request_buf);
            }
            else {
                HTTP_NewRequest(sess->server_fd,
                    sess->url->host, sess->url->full_loc,
                    sess->req.auth, sess->req.cookie,
					sess->req.nocache, sess->req.noprefetch);
            }
        }

        if(pipe_broken) {
            pipe_broken = 0;
            Fatal("ReachSession: pipe %d was broken.\n", sess->server_fd);
            LeaveSessionList(sess);
            release_socket(&(sess->server_fd));

            DeleteSession(sess);
        }
        else {
            sess->phase = PHASE_HEADER;
        }

    }
    else
    if(sess->req.method==METHOD_POST) {
        SendPostRequest(sess->server_fd, sess->url->path, sess->request_buf);

        sess->phase = PHASE_HEADER;
    }
    else {
        Error("ReachSession: strange method (%d)", sess->req.method);

        sess->phase = PHASE_ERROR;
    }

    return 0;
}


static sigjmp_buf request_jmp_point;

static RETSIGTYPE giveup_reception()
{
    SIGLONGJMP(request_jmp_point, 1);
}



int SwitchFTPHandler(int *fd, request_t req)
{
    pid_t cpid;
    int   cfd;

    cfd = *fd;
    cpid = fork();

    if(cpid<0) {
        Error("Cannot switch to FTPHandler");
        return -1;
    }
    if(cpid>0) {
        close(*fd);
        *fd = -1;

/*
    Trace("; pid# %d with fd# %d, cfd# %d, url '%s'", getpid(), *fd, cfd, xurl);
*/

        return 0;
    }
/*
    Trace("; pid# %d with fd# %d, cfd# %d, url '%s'", getpid(), *fd, cfd, xurl);
*/

#if 0
    Log("; FTP handler %d with fd# %d, url '%s'", getpid(), *fd, req.url);
#endif
    Trace("; FTP handler %d with fd# %d, url '%s'\n", getpid(), *fd, req.url);

    if(FTPHandler(req)) {
#if 0
        Log("; fail FTP access");
#endif
        exit(1);
    }

#if 0
    Log("; success FTP access");
#endif
    exit(0);

    return 0;
}

#if 0
int oldSwitchFTPHandler(int *fd, char *xurl)
{
    pid_t cpid;
    int   cfd;

    cfd = *fd;
    cpid = fork();

    if(cpid<0) {
        Error("Cannot switch to FTPHandler");
        return -1;
    }
    if(cpid>0) {
        close(*fd);
        *fd = -1;

    Trace("; pid# %d with fd# %d, cfd# %d, url '%s'", getpid(), *fd, cfd, xurl);

        return 0;
    }
    Trace("; pid# %d with fd# %d, cfd# %d, url '%s'", getpid(), *fd, cfd, xurl);

    Log("; FTP handler %d with fd# %d, url '%s'", getpid(), *fd, xurl);

    if(FTPHandler(cfd, xurl)) {
        Log("; fail FTP access");
        exit(1);
    }

    Log("; success FTP access");
    exit(0);

    return 0;
}
#endif



/*
 * do reception and generate new session
 *
 * returns:
 *      0   success
 *      1   error
 *
 */
int StartSession(Session *sess, int fd, int command_fd, request_t *orig_req)
{
    Info *info;
    int  chk;
    request_t req;
    hreport_t report;
    struct timeval send_end;
    int wlen;
    int qn;
    url_t *tmp_url;
    char rbuf[BUF_SIZE];

    /*
     */
    (*session_seqno)++;


    if(orig_req==NULL) {
        Trace(
  "StartSession: sess %p, client_fd# %d, command_fd# %d, orig_req *no*\n",
            sess, fd, command_fd);
    }
    else {
        Trace(
  "StartSession: sess %p, client_fd# %d, command_fd# %d, orig_req is ...\n",
            sess, fd, command_fd);
        ShowRequest(orig_req);
    }


#if 1
    /***
     *** remove noise of buggy sendmsg()/recvmsg()
     ***/
    if(orig_req!=NULL&&orig_req->mode==REQ_PREFETCH) {
        Trace(
  "StartSession: it is prefetch session, release client_fd# %d\n", fd);
        release_socket(&fd);
    }
#endif

    if(sess==NULL) {
        if((sess = NewSession())==NULL) {
            if(orig_req) {
                Error("StartSession: no idle session : %s '%s'",
                    ReqModeStr(orig_req->mode), orig_req->url);
            }
            else {
                Error("StartSession: no idle session : empty ");
            }
#if 0
            return 1;
#else

            if((sess = AltanativeSession())==NULL) {
                if(orig_req) {
                    Error("StartSession: no altanative session : %s '%s'",
                        ReqModeStr(orig_req->mode), orig_req->url);
                }
                else {
                    Error("StartSession: no altanative session : empty ");
                }

                return -1;
            }

        /*
            stop prefetch session
            and clean that session entity
        */


            return -1;

#endif
        }
    }
    else {
        if(sess->phase!=PHASE_IDLE) {
            Fatal("StartSession: specified session %p is not idle", sess);
        }
        sess->phase = PHASE_START;
    }

    sess->begin         = session_currenttime;
    sess->hot           = session_currenttime;
    sess->phase         = PHASE_REQUEST;
    sess->client_fd     = fd;
    sess->command_fd    = command_fd;


    /*
     * have request already. It is prefetch session, maybe.
     */
    if(fd==-1) {
        if(orig_req!=NULL) {
            Trace("StartSession: sid/id=%d/%d\n",
                orig_req->sid, orig_req->id);

            if(orig_req->mode==REQ_CLIENT) {
                Trace("StartSession: client request\n");

                req = *orig_req;

                tmp_url = URL_NewVal(orig_req->url);
            }
            else
            if(orig_req->mode==REQ_PREFETCH) {
                Trace("StartSession: prefetching request\n");

                req = *orig_req;

                tmp_url = URL_NewVal(orig_req->url);
            }
            else
            if(orig_req->mode==REQ_IMPORT) {
                Trace("StartSession: import request\n");

                req = *orig_req;

                tmp_url = URL_NewVal(orig_req->url);
            }
            else {
                Error("StartSession: fd is -1, orig_req is not NULL, mode %d",
                    orig_req->mode);
                return 1;
            }
        }
        else {
            Error("StartSession: fd is -1, but orig_req is NULL");

            DeleteSession(sess);
            return 1;
        }
    }

    /*
     * no request, let get it.
     */
    else {
        if(orig_req == NULL ||
          (orig_req != NULL && orig_req->mode==REQ_CLIENT)) {
            Trace("StartSession: client request\n");

            (*request_seqno)++;

            if(SIGSETJMP(request_jmp_point, 1)) {
                Error("StartSession: accept timeout #%d, give up.", fd);

                Log("; quiet connection req# %d, fd# %d, timeout.",
                    *request_seqno, fd);

#if 0
                close(fd);
                sess->client_fd = -1;
#endif
                DeleteSession(sess);

                return 1;
            }
            SetSignalHandler(SIGALRM, giveup_reception);
            alarm(reception_timeout);

            chk = HTTP_ParseRequest(fd, &req, sess->request_buf);
            SetSignalHandler(SIGALRM, SIG_IGN);

#ifdef TRACE_RECV_REQUEST

            dump_stringchunk("RECV", sess->request_buf);
#endif


            sess->hot = req.issue_time;

            if(chk == 1 || chk== -1) {
                Error("StartSession: ignore request from fd #%d (%d)\n", fd, chk);

                HTTP_ErrorMessage_Old(fd,
                    501, "Not implemented", "Unsupport Command", req.ver_major>0);
                release_socket(&fd);
                sess->client_fd = req.fd;

                report.type             = HREP_IGNOREREQUEST;
                if(orig_req!=NULL) {
                    report.sid          = orig_req->sid;
                    report.id           = orig_req->id;
                }
                report.keep_alive       = 0;

                ShowHReport(&report);
                XSendRecord(REPORT_TIMEOUT, command_fd, &report, sizeof(hreport_t));

                DeleteSession(sess);
                return 1;
            }
            else
            /* internal error */
            if(chk == -2 ) {
                Error("StartSession: ignore request from fd #%d (%d)\n", fd, chk);

                HTTP_ErrorMessage_Old(fd,
                    500, "Internal Error", "please wait several minites",
                    req.ver_major>0);
                release_socket(&fd);
                sess->client_fd = req.fd;

                report.type             = HREP_IGNOREREQUEST;
                if(orig_req!=NULL) {
                    report.sid          = orig_req->sid;
                    report.id           = orig_req->id;
                }
                report.keep_alive       = 0;

                ShowHReport(&report);
                XSendRecord(REPORT_TIMEOUT, command_fd, &report, sizeof(hreport_t));

                DeleteSession(sess);
                return 1;
            }


            if(orig_req!=NULL) {
                req.sid         = orig_req->sid;
                req.issue_time  = orig_req->issue_time;
            }

            tmp_url = URL_NewVal(req.url);
        }
        else {
            if(orig_req!=NULL) {
                Trace("StartSession: fd is not -1, but strange mode %d\n",
                    orig_req->mode);
            }
            else {
                Trace("StartSession: fd is not -1\n");
            }

#if 0
            Error("StartSession: ignore request from fd #%d\n", fd);
#endif
            Error("StartSession: ignore request, what is this ?\n");

            report.type             = HREP_IGNOREREQUEST;
            if(orig_req!=NULL) {
                report.sid          = orig_req->sid;
                report.id           = orig_req->id;
            }
            report.keep_alive       = 0;

            ShowHReport(&report);
            XSendRecord(REPORT_TIMEOUT, command_fd, &report, sizeof(hreport_t));

            DeleteSession(sess);
            return 1;
        }
    }


    /*
     * Acceptable Request is arrive.
     */

#if 0
    tmp_url = URL_NewVal(req.url);
#endif
    URL_Normalize(tmp_url);
    URL_Show(tmp_url);
    strcpy(req.url, tmp_url->full_loc);

#if 0
    sess->type  = req.mode;
#endif

    switch(req.mode) {
    case REQ_CLIENT:    sess->type = SESS_CLIENT;   break;
    case REQ_PREFETCH:  sess->type = SESS_PREFETCH; break;
    case REQ_IMPORT:    sess->type = SESS_IMPORT;   break;
    defauit:
                        Log("; strange request (%d)", req.mode);
                        sess->type = SESS_NOP;      break;
    }

    sess->req   = req;
    sess->url   = tmp_url;
    strcpy(sess->_url, tmp_url->full_loc);

    ShowRequest(&req);

#if 0
    Trace("_URL: 0: %p '%s'\n", sess, sess->_url);
#endif


    /*
     * choice send or fetch
     */
    sess->phase = PHASE_THINK;


    switch(req.method) {
    case METHOD_POST:
#if 0
        strcpy(sess->body_buf, rbuf);
#endif

        if(!ConnectSession(sess)) {
            JoinSessionList(sess);
        }

        break;

    case METHOD_GET:
        if(strncmp_first_all(req.url, 3, 'f', "ftp")) {

#if 0
            Log("; FTP session for '%s'", req.url);
#endif
            Trace("; FTP session for '%s'\n", req.url);

#if 0
            SwitchFTPHandler(&sess->client_fd, req.url);
#endif
            SwitchFTPHandler(&sess->client_fd, req);

            URL_Delete(tmp_url);
            tmp_url = NULL;

            /*
            sess->client_fd = -1;
            release_socket(&sess->client_fd);
            */
            DeleteSession(sess);

#if 0
            report.type         = HREP_CACHEHIT;
            report.sid          = req.sid;
            report.id           = req.id;
            report.keep_alive   = 0;
            report.req          = req;

            /*
             * To avoid deadlock with IPC, set timer
             */
            ShowHReport(&report);
            XSendRecord(REPORT_TIMEOUT, command_fd, &report, sizeof(hreport_t));
#endif


            break;

        }
        /*
         * Send Direct
         */
        if( !req.nocache && (info=Have(req.url))!=NULL) {

            if(sess->type==SESS_PREFETCH && req.hereport==-1) {
                Log("; got prefetch command, but I have it ('%s')", req.url);
            }
            else {
#if 1
                Trace("StartSession: have '%s'\n", req.url);
#endif
                sess->phase = PHASE_SEND;

                wlen = Send(&fd, req.url, info, req.ver_major, req.ver_minor, 0, 0);
                release_socket(&fd);
                sess->client_fd = fd;   /* avoid close at DeleteSession() */

                gettimeofday(&send_end, NULL);

                /*
                 * SEND url head-size body-size
                 */
                Log("%d.%06d SEND %d %s HTTP %d.%d %s %s %s %s %s %d %s %ld %d : %s %s %d %ld %d : %s %d",
                    send_end.tv_sec, send_end.tv_usec,
                    req.hereport, req.peername,
                    req.ver_major, req.ver_minor,
                    req.nocache == 1 ? "n" : "-",
                    req.ims == -1 ? "-" : "*",
    /*
                    req.auth[0] == '\0' ? "-" : req.auth,
    */
                    req.auth[0] == '\0' ? "-" : "X",
                    ReqMethodStr(req.method), req.url,
                    info->attr.state,
                    info->attr.type,
                    (long)wlen,
                    WCOL_SUB_TIME(send_end, req.issue_time),
                    "CACHE", "-", -1,
                    (long)info->attr.length, info->attr.take_time,
                    info->attr.init_state==REQ_CLIENT ?       "CLIENT  " :
                      (info->attr.init_state==REQ_PREFETCH ?  "PREFETCH" :
                        (info->attr.init_state==REQ_IMPORT ?  "IMPORT  " :
                                                              "UNKNOW  ")
                      ),
                    info->attr.count+1
                );
            }


            URL_Delete(tmp_url);
            tmp_url = NULL;

            DeleteSession(sess);

            report.type         = HREP_CACHEHIT;
            report.sid          = req.sid;
            report.id           = req.id;
            report.keep_alive   = 0;
            report.req          = req;

            /*
             * To avoid deadlock with IPC, set timer
             */
            ShowHReport(&report);
            XSendRecord(REPORT_TIMEOUT, command_fd, &report, sizeof(hreport_t));


        }
        else {
#if 1
            Trace("StartSession: *not* have '%s'\n", req.url);
#endif



#if 0
            /* lock session, if same URL is retrieved */
            if(IsInSesssionPool(tmp_url->full_loc, sess)) {
                sess->phase = PHASE_WAIT;
                return 0;
            }
#endif


            if(req.nocache==0 && num_sibling>0) {
                sess->phase  = PHASE_CONNECTSIBLING;
                qn =  ConnectQuerySiblingSession(sess);

#if 1
                if(sess->query_sibling>0) {
                    Trace("StartSession: quering %d host(s)\n",
                        sess->query_sibling);
                }
#endif

                if(sess->query_sibling<1) {
                    if(!ConnectSession(sess)) {
                        JoinSessionList(sess);
                    }
                }
            }
            else {
                if(!ConnectSession(sess)) {
                    JoinSessionList(sess);
                }
            }

#if 0
    Trace("_URL: 1: %p '%s'\n", sess, sess->_url);
#endif

        }

        break;

    default:
        Error("StartSession: Sorry, Unsupport method '%s'.\n",
            ReqMethodStr(req.method));

#if 0
        HTTP_ErrorMessage_Old(fd,
            501, "Not implemented", "Unsupport method", req.ver_major>0);
#endif
        HTTP_ErrorMessage(fd, req.ver_major, req.ver_minor,
            501, "Not implemented",
            "Unsupport method (%s)", ReqMethodStr(req.method));

        release_socket(&fd);
        sess->client_fd = req.fd;

        report.type             = HREP_IGNOREREQUEST;
        if(orig_req!=NULL) {
            report.sid          = orig_req->sid;
            report.id           = orig_req->id;
        }
        report.keep_alive       = 0;

        ShowHReport(&report);
        XSendRecord(REPORT_TIMEOUT, command_fd, &report, sizeof(hreport_t));

        DeleteSession(sess);

        return 1;
    }

    return 0;
}






static int StoreHeader(Session *sess)
{
    char tmp[STRING_SIZE];

    if(SetupFile(sess->header_fn, sess->url->full_loc, HEAD_TMP_FILE,
             &(sess->header_fd))) {
        return 1;
    }

    if(sess->hsum > HEADER_BUFLEN) {
        Log("; StoreHeader: %p - header over run, length %ld/%ld",
            sess, (long)sess->hsum, (long)HEADER_BUFLEN);
    }

w1:
    if(lseek(sess->header_fd, (off_t) 0, SEEK_SET)<0) {
        if(errno==EINTR)
            goto w1;
        Error("StoreHeader: fail write() (%d)\n", errno);
    }
    if(write(sess->header_fd, sess->header_buf, sess->hsum)<0) {
        if(errno==EINTR)
            goto w1;
        Error("StoreHeader: fail write() (%d)\n", errno);
    }
#ifdef TRACE_BUF
    dump_stringchunk("  header", sess->header_buf);
#endif

f1:
    if(ftruncate(sess->header_fd, sess->hsum)) {
        if(errno==EINTR)
            goto f1;
        Error("StoreHeader: fail ftruncate(), sess %p, header_fd# %d (%d)",
            sess, sess->header_fd,  errno);
    }

    if(close(sess->header_fd)) {
        Error("StoreHeader: fail close(), sess %p, header_fd# %d (%d)",
            sess, sess->header_fd, errno);
    }
    sess->header_fd = -1;

    AssignFileName(tmp, sess->url->full_loc, HEAD_FILE);

r1:
    if(rename(sess->header_fn, tmp)) {
        if(errno==EINTR)
            goto r1;
   Error("StoreHeader: fail rename(), sess %p, header_fn '%s', tmp '%s' (%d)\n",
            sess, sess->header_fn, tmp, errno);
        if(errno==ENOENT) {
            Trace(" File is not exist\n TEMP FILE %s\n ORIG FILE %s\n",
                sess->header_fn, tmp);
        }
    }
    else {
        strcpy(sess->header_fn, tmp);
    }

    return 0;
}



static int StoreBody(Session *sess)
{
    char tmp[STRING_SIZE];
    char tmp1[STRING_SIZE], tmp2[STRING_SIZE];
    char *conv_byte_real(off_t);

#if 1

#if 0
    if(SetupFile(sess->body_fn, sess->url->full_loc, BODY_TMP_FILE,
            &(sess->body_fd))) {
        return 1;
    }
#endif

    if(sess->bsum > BODY_BUFLEN) {
        /*
        Log("; StoreBody:   %p - body   over run, length %ld/%ld",
            sess, (long)sess->bsum, (long)BODY_BUFLEN);
        */
#if 1
        strcpy(tmp1, conv_byte_real(sess->bsum));
        strcpy(tmp2, conv_byte_real((off_t)BODY_BUFLEN));
        Log("; StoreBody:   %p - body   over run, length %ld (%s) / %ld (%s)",
            sess, (long)sess->bsum, tmp1, (long)BODY_BUFLEN, tmp2);
#endif
    }

f1:
    if(ftruncate(sess->body_fd, sess->bsum)) {
        if(errno==EINTR)
            goto f1;
        Error("StoreBody: fail ftruncate(), sess %p, body_fd# %d (%d)",
            sess, sess->body_fd, errno);
    }
    if(close(sess->body_fd)) {
        Error("StoreBody: fail close(), sess %p, body_fd# %d (%d)",
            sess, sess->body_fd, errno);
    }
    sess->body_fd = -1;

    AssignFileName(tmp, sess->url->full_loc, BODY_FILE);
r1:
    if(rename(sess->body_fn, tmp)) {
        if(errno==EINTR)
            goto r1;
      Error("StoreBody: fail rename(), sess %p, body_fn '%s', tmp '%s' (%d)\n",
            sess, sess->body_fn, tmp, errno);
        if(errno==ENOENT) {
            Trace(" File is not exist\n TEMP FILE %s\n ORIG FILE %s\n",
                sess->body_fn, tmp);
        }
        if(errno==EISDIR) {
            Trace(" File is directory\n TEMP FILE %s\n ORIG FILE(DIR) %s\n",
                sess->body_fn, tmp);
        }
    }
    else {
        strcpy(sess->body_fn, tmp);
    }
#endif

    return 0;
}





/*
 *
 */
int CloseSession(Session *sess)
{
    char tmp[STRING_SIZE];
    long conn, trans, takes;
    char info_type[STRING_SIZE];
    int  _chk_;
    hreport_t report;

    report.type = HREP_UNKNOW;
    sess->phase = PHASE_END;


    gettimeofday(&(sess->end_tp), NULL);



    /*
     * Connections
     */
    Trace("CloseSession: END #%d, len %d, type %d\n",
        sess->server_fd, sess->hsum+sess->bsum, sess->type);

    LeaveSessionList(sess);
    release_socket(&(sess->server_fd));

/*
    sess->wlen += sess->bsum;
*/


    /*
     * store data and information
     */

    conn  = SESSION_SUB_TIME(sess, conn_tp, start_tp);
    trans = SESSION_SUB_TIME(sess, end_tp, conn_tp);
    takes = SESSION_SUB_TIME(sess, end_tp, start_tp);


    if(conn<0 || trans<0 || takes<0) {
        fprintf(stdout, "\tSTART %9d %9d\n\tCONN  %9d %9d\n\tTAKES %9d %9d\n",
            sess->start_tp.tv_sec,  sess->start_tp.tv_usec,
            sess->conn_tp.tv_sec,   sess->conn_tp.tv_usec,
            sess->end_tp.tv_sec,    sess->end_tp.tv_usec);
    }

#if 1
    Trace("CloseSession: URL '%s'\n", sess->url->full_loc);
#endif

    strcpy(sess->info.attr.name, sess->url->full_loc);
    sess->info.attr.hsize       = sess->hsum;
    sess->info.attr.bsize       = sess->bsum;
    sess->info.attr.length      = sess->hsum + sess->bsum;
    sess->info.attr.last        = (time_t)((sess->conn_tp).tv_sec);
    sess->info.attr.conn_time   = conn;
    sess->info.attr.trans_time  = trans;
    sess->info.attr.take_time   = takes;

    if(sess->type==SESS_CLIENT) {
        sess->info.attr.init_state  = REQ_CLIENT;
        sess->info.attr.count       = 1;
    }
    if(sess->type==SESS_PREFETCH) {
        sess->info.attr.init_state  = REQ_PREFETCH;
        sess->info.attr.count       = 0;
    }
    if(sess->type==SESS_IMPORT) {
        sess->info.attr.init_state  = REQ_IMPORT;
        sess->info.attr.count       = 0;
    }

#if 1
    if(trace_flag)
        What(&(sess->info));
#endif
    StoreHeader(sess);
    StoreBody(sess);

    if(SaveInfo(&(sess->info), sess->url->full_loc)) {
        Error("CloseSession: cannot save info");
    }

    if(sess->bsum==0) {
        Error("CloseSession: Empty '%s'", sess->url->full_loc);
    }


    /*
     * report
     */
    if(sess->type==SESS_CLIENT) {
        Log("%d.%06d FETCH %d %s HTTP %d.%d %s %s %s %s %s %d %s %ld %d : %s %s %d %ld %d",
            sess->end_tp.tv_sec, sess->end_tp.tv_usec,

            sess->req.hereport, sess->req.peername,
            sess->req.ver_major, sess->req.ver_minor,
            sess->req.nocache == 1 ? "n" : "-",
            sess->req.ims == -1 ? "-" : "*",
/*
            sess->req.auth[0] == '\0' ? "-" : sess->req.auth,
*/
            sess->req.auth[0] == '\0' ? "-" : "X",

            ReqMethodStr(sess->req.method), sess->url->full_loc,
            sess->info.attr.state,
            sess->info.attr.type,
            (long)(sess->wlen),
            WCOL_SUB_TIME(sess->end_tp, sess->req.issue_time),

            ServerTypeStr(sess->server_type),
            sess->server_name, sess->server_port,
            (long)(sess->hsum+sess->bsum), takes);
    }
    else
    if(sess->type==SESS_PREFETCH) {
        Log("%d.%06d PREFETCH - - - - - - - %s %s %d %s %ld %d : %s %s %d %ld %d",
            sess->end_tp.tv_sec, sess->end_tp.tv_usec,

            ReqMethodStr(sess->req.method), sess->url->full_loc,
            sess->info.attr.state,
            sess->info.attr.type[0] == '\0' ? "-" : sess->info.attr.type,
            (long)sess->wlen,
            WCOL_SUB_TIME(sess->end_tp, sess->req.issue_time),

            ServerTypeStr(sess->server_type),
            sess->server_name, sess->server_port,
            (long)(sess->hsum+sess->bsum), takes);
    }

    report.type         = HREP_SERVERACCESS;
    report.sid          = sess->req.sid;
    report.id           = sess->req.id;
    report.keep_alive   = 0;
    report.req          = sess->req;

    if(trace_flag) {
        Trace("handler: send report which following...\n");
        ShowHReport(&report);
    }

    /*
     * To avoid deadlock with IPC, set timer
     */

    pipe_broken = 0;
    XSendRecord(REPORT_TIMEOUT, sess->command_fd, &report, sizeof(hreport_t));

    DeleteSession(sess);

    return 0;
}



/*
 *
 */
int AbortSession(Session *sess)
{
    /*
    int info_fd, count;
    char *info_path;
    */
    char *tmp;
    long conn, trans, takes;
    char info_type[STRING_SIZE];
    int _chk_;

    Trace("AbortSession: URL '%s', h %d, b %d\n",
        sess->url->full_loc, sess->hsum, sess->bsum);

    ftruncate(sess->body_fd, sess->bsum);

    close(sess->header_fd);
    sess->header_fd = -1;
    close(sess->body_fd);
    sess->body_fd = -1;


    /*
     * Connections
     */
    LeaveSessionList(sess);
    release_socket(&(sess->server_fd));

    Trace("AbortSession: END #%d, len %d, type %d\n",
        sess->server_fd, sess->hsum+sess->bsum, sess->type);

    DeleteSession(sess);

    return 0;
}


int HTTP_ParseHeaderBuffer(Info *info, char *hbuf);

/*
 * ReadSession
 *
 *
 *  returns
 *      -1  error
 *       0  success
 *       1  success and end of data
 */
int ReadSession(Session *sess)
{
    char *tmp;
    int  count;
    long conn, trans, takes;
    int  pos = 0;
    int  server_state = -1;
    int  nwrite;



#ifdef TRACE_BUF
    Trace("ReadSession: fd=%d, phase=%d, mode=%d.\n",
        sess->server_fd, sess->phase, sess->rcv_mode);
#endif

    if(sess==NULL) {
        return -1;
    }

#if 0
    sess->hot = session_currenttime;
#endif

    memset(sess->linebuf, 0, BUF_SIZE);
    if((count = read(sess->server_fd, sess->linebuf, BUF_SIZE))<=0) {
        /*
         * ERROR
         */
        if(count<0) {
            if(errno == EINTR) {        /* intrrupt */
                return 0;
            }
            if(errno == EAGAIN) {       /* empty */
                return 0;
            }

            if(errno == ECONNRESET) {   /* connection reset by peer */
                Error("ReadSession: connection(%d) reset by peer",
                    sess->server_fd);

                return -1;
            }
            Error("ReadSession: read error on fd #%d (errno=%d)\n",
                sess->server_fd, errno);
            return -1;
        }

        /*
         * END  count==0
         */
        return 1;
    }

    sess->hot = session_currenttime;

#ifdef TRACE_BUF
    Trace("ReadSession: fd# %3d, %c, read %7d, mode %d ",
        sess->server_fd, sess->phase==PHASE_HEADER ? 'H' : '?',
        count, sess->rcv_mode);
    dump_string("", sess->linebuf);
    dump_stringchunk("  linebuf   ", sess->linebuf);
    dump_stringchunk("  header_buf", sess->header_buf);
#endif


    if(sess->phase==PHASE_HEADER) {
        int  i, len;
        char *p, *q;
        char wbuf[STRING_SIZE];


        if(sess->hsum+sess->bsum==0) {
            sess->rcv_mode = 0;

            if(count<14) {
                Trace("ReadSession: @@ - fd %d, read %d [%s]\n",
                    sess->server_fd, count, sess->linebuf);

                sess->phase = PHASE_BODY;
                pos = 0;

                goto RCV_BODY;
            }

            strncpy(wbuf, sess->linebuf, 14);
            wbuf[14] = '\0';
#ifdef TRACE_BUF
            Trace("ReadSession: @@ + fd %d, read %d [%s]\n",
                sess->server_fd, count, wbuf);
#endif

            if(strncmp(sess->linebuf,"HTTP/",5)!=0) {
                sess->phase = PHASE_BODY;
                pos = 0;

                goto RCV_BODY;
            }


            p = &sess->linebuf[5];

            q = &wbuf[0];
            while(*p && isdigit(*p))
                *q++ = *p++;
            *q = '\0';
#ifdef TRACE_BUF
            Trace("ReadSession: wbuf -> [%s]\n", wbuf);
#endif
            sess->info.attr.ver_major = atoi(wbuf);
            sess->info.attr.ver_minor = 0;


            if(*p=='.') {
                p++;

                q = &wbuf[0];
                while(*p && isdigit(*p))
                    *q++ = *p++;
                *q = '\0';
#ifdef TRACE_BUF
                Trace("ReadSession: wbuf -> [%s]\n", wbuf);
#endif
                sess->info.attr.ver_minor = atoi(wbuf);
            }

            if(*p==' ') {
                p++;

                q = &wbuf[0];
                while(*p && isdigit(*p))
                    *q++ = *p++;
                *q = '\0';
#if 0
                Trace("ReadSession: wbuf -> [%s]\n", wbuf);
#endif
                sess->info.attr.state = atoi(wbuf);
            }

#ifdef TRACE_BUF
            Trace("ReadSession: ver %d.%d, code %d\n",
                sess->info.attr.ver_major,
                sess->info.attr.ver_minor,
                sess->info.attr.state);
#endif

        }


        len = count;

        /*
         * rcv_mode:
         *  LF...    none plain     \n \n
         *            -1    0       2  4
         *  CRLF...  none plain  \r \n \r \n
         *            -1    0    1  2  3  4
         *
         */

        if(sess->rcv_mode==3) {
            if(count>=1 && sess->linebuf[0]=='\n') {
                pos = 1;
                len = 1;
                sess->rcv_mode = 4;
            }
        }

        if(sess->rcv_mode==2) {
            if(count>=1 && sess->linebuf[0]=='\n') {
                pos = 1;
                len = 1;
                sess->rcv_mode = 4;
            }
            else
            if(count>=2 && sess->linebuf[0]=='\r' && sess->linebuf[1]=='\n') {
                pos = 2;
                len = 2;
                sess->rcv_mode = 4;
            }
            else
            if(count==1 && sess->linebuf[0]=='\r') {
                pos = 0;
                len = 1;
                sess->rcv_mode = 3;
            }
            else {
                sess->rcv_mode = 0;
            }
        }

        if(sess->rcv_mode==1) {
            if(count>=3 && sess->linebuf[0]=='\n'
                    && sess->linebuf[1]=='\r' && sess->linebuf[2]=='\n') {
                pos = 3;
                len = 3;
                sess->rcv_mode = 4;
            }
            else {
                sess->rcv_mode = 0;
            }
        }


        if(sess->rcv_mode==0) {
#ifdef TRACE_BUF
    Trace("Current mode 0\n");
#endif
            /*
             * search header end.
             */
            i = 0;
            while(i<count) {
                while(i<count&& sess->linebuf[i]!='\n')
                    i++;

                if(sess->linebuf[i]=='\n') {
                    sess->rcv_mode = 2;
#ifdef TRACE_BUF
    Trace("Eneter mode 2\n");
#endif

                    /*
                     * for LF splited header like NCSA httpd
                     */
                    if(i+1<count
                        && sess->linebuf[i+1]=='\n') {
                        pos = i+2;
                        len = i+2;
                        sess->rcv_mode = 4;
#ifdef TRACE_BUF
    Trace("Eneter mode 4\n");
#endif
                        break;
                    }
                    else
                    /*
                     * for CRLF splited header like CERN httpd
                     */
                    if(i+2<count
                        && sess->linebuf[i+1]=='\r' && sess->linebuf[i+2]=='\n') {
                        pos = i+3;
                        len = i+3;
                        sess->rcv_mode = 4;
#ifdef TRACE_BUF
    Trace("Eneter mode 4\n");
#endif
                        break;
                    }

                    i++;
                }
                else {
                    while(i<count&& sess->linebuf[i]!='\r')
                        i++;

                    if(sess->linebuf[i]=='\r') {
                        sess->rcv_mode = 1;

#ifdef TRACE_BUF
    Trace("Eneter mode 1\n");
#endif

                        /*
                         * for CRLF splited header like CERN httpd
                         */
                        if(i+3<count
                            && sess->linebuf[i+1]=='\n'
                            && sess->linebuf[i+2]=='\r' && sess->linebuf[i+3]=='\n') {
                            pos = i+4;
                            len = i+4;
                            sess->rcv_mode = 4;
#ifdef TRACE_BUF
    Trace("Eneter mode 4\n");
#endif
                            break;
                        }

                        i++;
                    }
                    else {
                        sess->rcv_mode = 0;
#ifdef TRACE_BUF
    Trace("Eneter mode 0\n");
#endif
                        i++;
                    }
                }
            }

#if 0
            Trace("\ti/count=%d/%d\n",i,count);
#endif
        }

#ifdef TRACE_BUF
        Trace("ReadSession: fd# %3d, read %7d, header %7d, %7d..%7d %d\n",
            sess->server_fd, count,
            len, sess->hsum, sess->hsum+len,
            sess->rcv_mode
            );
#endif

        /* strcat(sess->header_buf, sess->linebuf) */
        {
            char *p, *q;
            int c;

            p = &sess->linebuf[0];
            q = &sess->header_buf[sess->hsum];
            c = 0;
            while(c<len) {
                *q++ = *p++;
                c++;
            }
            *q = '\0';
        }


#ifdef TRACE_BUF
        dump_stringchunk("  header", sess->header_buf);
#endif


        if(sess->rcv_mode==4) {
            sess->phase=PHASE_BODY;

            if(sess->client_fd>=0) {
                nwrite = SendHeadBuf(&sess->client_fd, sess->header_buf,
                    sess->url->full_loc, NULL,
                    sess->req.ver_major, sess->req.ver_minor, 0, 0);
                if(nwrite>=0) {
                    sess->wlen += nwrite;
                }
            }

            HTTP_ParseHeaderBuffer(&(sess->info), sess->header_buf);
#if 0
            Trace("-=-=-=-=-=-=-=-\n");
            if(trace_flag)
                What(&sess->info);
#endif
        }

        sess->hsum += len;
    }


RCV_BODY:




    /* save rest of buffer as body */
    if(sess->phase==PHASE_BODY) {
#if 0
        Trace("ReadSession: write body %d bytes (from %d).\n",
            count-pos, pos);
#endif

#if 0
    if(sess->bsum+count > BODY_BUFLEN) {
        Fatal("ReadSession: %p - body over run, count %d, body %d/%d",
            sess, count, sess->bsum, BODY_BUFLEN);
    }
#endif

        if(count-pos>0) {

            if(sess->body_fd != -1) {
write_bfile:
                if((nwrite = write(sess->body_fd, &sess->linebuf[pos], count-pos))<0) {
                    if(errno==EINTR) {
                        goto write_bfile;
                    }
                    Trace("ReadSession: write() error body_fd fd# %d (%d)\n",
                        sess->body_fd, errno);

                    sess->body_fd = -1;
                }
                else {
                    sess->bsum += nwrite;
                }
            }

            if(sess->client_fd != -1) {
write_sock:
                pipe_broken = 0;
                if((nwrite = write(sess->client_fd,
                        &sess->linebuf[pos], count-pos))<=0) {
                    if(errno==EINTR) {
                        goto write_sock;
                    }
                    if(pipe_broken) {
                        Trace("\tpipe broken\n");
                        pipe_broken = 0;
                    }

                    Trace("ReadSession: write() error client_fd fd# %d (%d)\n",
                        sess->client_fd, errno);

                    release_socket(&(sess->client_fd));
                }
                else {
                    sess->wlen += nwrite;
                }
            }
        }
#if 0
        else {
            Trace("ReadSession: no body, thru\n");
        }
#endif


    }

#ifndef NO_CLENGTH_CHECK
    /*
    Trace("ReadSession: fd# %2d, clen %d, bsum %d\n",
        sess->server_fd, sess->info.attr.clength, sess->bsum);
    */

    if(sess->info.attr.clength >=0 &&
            sess->bsum>=sess->info.attr.clength) {
        Trace("ReadSession: fd# %2d, over\n",
            sess->server_fd);
        return 1;   /* success (over) */
    }
    /*
    Trace("ReadSession: fd# %2d, continued...\n",
        sess->server_fd);
    */
#endif

    return 0;   /* success (continued) */
}



/*
 *
 */
OneConnectSiblingSession(Session *sess,
        char *hostname, struct sockaddr_in *hostaddr, int portno)
{
    int fd;

    Trace("OneConnectSiblingSession:\n");

    if(hostname!=NULL) {
        if((fd=UDP_aconnect(hostname, portno))<0) {
            Error("OneConnectSiblingSession: cannot connect sibling name");
            return -1;
        }
    }
    else
    if(hostaddr!=NULL) {
        if((fd=UDP_aconnect_addr(hostaddr, portno))<0) {
            Error("OneConnectSiblingSession: cannot connect sibling addr");
            return -1;
        }
    }
    else {
        return -1;
    }

    return 0;
}



/*
 *
 */
int OneConnectQuerySession(Session *sess,
        char *hostname, struct sockaddr_in *hostaddr, int portno)
{
    int fd;
    struct sockaddr_in addr;


    if(hostaddr == NULL) {
        Trace("OneConnectQuerySession: host %s, port %d\n",
            hostname, portno);
        if(find_addr(hostname, portno, &addr)) {
            Error("OneConnectQuerySession:Cannot found host '%s'", hostname);
            return -1;
        }
        Trace("\tgot address\n");
    }
    else {
        Trace("OneConnectQuerySession: addr -, port %d\n", portno);
    }

    if((fd = socket(AF_INET, SOCK_DGRAM, 0))<0) {
        Error("OneConnectQuerySession: cannot create socket");
        return -1;
    }
    Trace("OneConnectQuerySession: socket %d\n", fd);

    if(ICP_SendToQuery(fd, &addr, sizeof(struct sockaddr_in),
        sess->req.sid, sess->url->full_loc)<=0) {

        close(fd);
        return -1;
    }
    else {
        return fd;
    }

    /*
    return -1;
    */
}


int OneConnectNotifySession(Session *sess,
        char *hostname, struct sockaddr_in *hostaddr, int portno)
{
    int fd;
    struct sockaddr_in addr;


    if(hostaddr == NULL) {
        Trace("OneConnectNotifySession: name %s\n", hostname);
        if(find_addr(hostname, portno, &addr)) {
            Error("OneConnectNotifySession:Cannot found host '%s'", hostname);
            return -1;
        }
    }
    else {
        Trace("OneConnectNotifySession: addr -\n");
    }

    if((fd = socket(AF_INET, SOCK_DGRAM, 0))<0) {
        Error("OneConnectNotifySession: cannot create socket");
        return -1;
    }
    if(ICP_SendToNotify(fd, &addr, sizeof(struct sockaddr_in),
        sess->req.sid, sess->url->full_loc)<=0) {

        close(fd);
        return -1;
    }
    else {
        return fd;
    }

    /*
    return -1;
    */
}



/*
 *
 */
int ConnectQuerySiblingSession(Session *sess)
{
    int i;
    int icp_fd;
    int c;

    Trace("ConnectQuerySiblingSession: have %d sibling(s)\n", num_sibling);

    c = 0;
    for(i=0;i<num_sibling;i++) {
        if(sibling_list[i]==NULL) {
            Trace("sibling %2d *empty*\n", i);
            continue;
        }

        if(sibling_list[i]->type & SIBLING_TYPE_QUERY) {

            Trace("Query  sibling# %2d %s\n", i, sibling_list[i]->name);

            if((icp_fd = OneConnectQuerySession(sess,
                            sibling_list[i]->name, NULL,
                            sibling_list[i]->icp_port))<0) {
                Trace("ConnectQuerySiblingSession: fail to query\n");
            }
            else {
                Trace("ConnectQuerySiblingSession: join %d new query\n", icp_fd);

                if(icp_fd>last_fd) {
                    last_fd = icp_fd;
                }

                FD_SET(icp_fd, &session_fdset);
                fd2sess[icp_fd].type            = CONN_ICPQUERY;
                fd2sess[icp_fd].sess            = sess;
                fd2sess[icp_fd].sibling_id      = sibling_list[i]->id;
                fd2sess[icp_fd].timeout         = session_currenttime;
                WCOL_ADD_MSEC(fd2sess[icp_fd].timeout, query_timeout);
#if 0
                Trace("*** diff %d/%d\n", WCOL_SUB_TIME(fd2sess[icp_fd].timeout, session_currenttime), query_timeout);
#endif

                sess->query_sibling++;
                c++;
            }
        }

        if(sibling_list[i]->type & SIBLING_TYPE_NOTIFY) {

            Trace("Notify sibling# %2d %s\n", i, sibling_list[i]->name);

            if((icp_fd = OneConnectNotifySession(sess,
                            sibling_list[i]->name, NULL,
                            sibling_list[i]->icp_port))<0) {
                Trace("ConnectNotifySiblingSession: fail to notify\n");
            }
            else {
                Trace("ConnectNotifySiblingSession: notify is over fd# %d\n",
                    icp_fd);
                close(icp_fd);

#if 0
                Trace("ConnectNotifySiblingSession: join %d new notify\n", icp_fd);

                if(icp_fd>last_fd) {
                    last_fd = icp_fd;
                }

                FD_SET(icp_fd, &session_fdset);
                fd2sess[icp_fd].type            = CONN_ICPQUERY;
                fd2sess[icp_fd].sess            = sess;
                fd2sess[icp_fd].sibling_id      = sibling_list[i]->id;
                fd2sess[icp_fd].timeout         = session_currenttime;
                WCOL_ADD_MSEC(fd2sess[icp_fd].timeout, query_timeout);
#if 1
                Trace("*** diff %d/%d\n", WCOL_SUB_TIME(fd2sess[icp_fd].timeout, session_currenttime), query_timeout);
#endif

                sess->query_sibling++;
                c++;
#endif
            }
        }

    }
    sess->phase = PHASE_QUERY;

#if 0
    DisplaySessions(0);
#endif

    return c;
}



/*
 *
 */
int ConnectSiblingSession(Session *sess)
{
    int i;
    int icp_fd;

    Trace("ConnectSiblingSession:\n");

    for(i=0;i<1;i++) {
        if((icp_fd = OneConnectSiblingSession(sess, "cache1.nara.wide.ad.jp", NULL, 3130))<0) {
        }
        else {
            Trace("ConnectSiblingSession: join new query\n");

            if(icp_fd>last_fd) {
                last_fd = icp_fd;
            }

            FD_SET(icp_fd, &session_fdset);
            fd2sess[icp_fd].type = CONN_ICPCONNECT;
            fd2sess[icp_fd].sess = sess;
            fd2sess[icp_fd].timeout = session_currenttime;
            fd2sess[icp_fd].timeout.tv_sec += 1;

            sess->query_sibling++;
        }
    }
    sess->phase = PHASE_CONNECTSIBLING;

#if 0
    DisplaySessions(0);
#endif

    return 0;
}


/*
 * QuerySession
 */
int QuerySession(Session *sess, int icp_fd)
{
    sess->phase = PHASE_QUERY;

    Trace("QuerySession: send request to sibling '%s' icp_fd %d\n",
            sess->url->full_loc, icp_fd);
    ICP_SendQuery(icp_fd, sess->req.sid, sess->url->full_loc);
#if 1
    fflush(stdout);
    fflush(stderr);
#endif

    return 0;
}


/*
 * ClearQuerySession
 */
int ClearQuerySession(int icp_fd)
{
    Trace("ClearQuerySession: fd# %d\n", icp_fd);

    FD_CLR(icp_fd, &session_fdset);
    fd2sess[icp_fd].type = CONN_NOP;
    fd2sess[icp_fd].sess = NULL;

    release_socket(&icp_fd);

    CheckLastFd();

    return 0;
}


sibling_info_t *find_sibling(int);


/*
 * ReceiveQueryReplySession
 */
int ReceiveQueryReplySession(Session *sess, int icp_fd)
{
    char icp_msg[BUF_SIZE];
    int nread;
    sibling_info_t *target_sibling;

    Trace("ReceiveQueryReplySession: sess %8x, icp_fd %d\n", sess, icp_fd);

    FD_CLR(icp_fd, &session_fdset);
    target_sibling = find_sibling(fd2sess[icp_fd].sibling_id);

    fd2sess[icp_fd].type        = CONN_NOP;
    fd2sess[icp_fd].sess        = NULL;
    fd2sess[icp_fd].sibling_id  = -1;

    if((nread = read(icp_fd, icp_msg, BUF_SIZE))<0) {
        Error("ReceiveQueryReplySession: cannot read fd=%d", icp_fd);
    }
    release_socket(&icp_fd);
    sess->query_sibling--;

    CheckLastFd();

#if 1
    ICP_DumpMessage(stderr, (unsigned char*)icp_msg);
    fflush(stderr);
    fflush(stdout);
#endif

    /* already found */
    if(sess->phase==PHASE_CONNECTSERVER) {
        return 0;
    }
    if(sess->phase!=PHASE_QUERY) {
        Log("; stringe phase %s , sess %p", SessPhaseStr(sess->phase), sess);
        return 0;
    }

    if(target_sibling!=NULL && ICP_IsHit(icp_msg)) {
        Trace("ReceiveQueryReplySession: resouce is found in '%s'\n",
            target_sibling->name);
        if(!ConnectSession_Sibling(sess, target_sibling)) {
            JoinSessionList(sess);

            return 0;
        }

        return -1;
    }


    if(sess->query_sibling==0) {
        if(!ConnectSession(sess)) {
            JoinSessionList(sess);

            return 0;
        }

        return -1;
    }

    return 0;
}


/*
 * RemoveUnreachQuerySession -
 */
int RemoveUnreachQuerySession(Session *sess, int icp_fd)
{
    Trace("RemoveUnreachQuerySession: unreach query, try to connect origin server\n");

    FD_CLR(icp_fd, &session_fdset);
    fd2sess[icp_fd].type = CONN_NOP;
    fd2sess[icp_fd].sess = NULL;

    release_socket(&icp_fd);
    sess->query_sibling--;

    CheckLastFd();

    if(sess->phase!=PHASE_QUERY) {
        Log("; stringe phase %s , sess %p", SessPhaseStr(sess->phase), sess);
        return 0;
    }

    if(sess->query_sibling==0) {
        if(!ConnectSession(sess)) {
            JoinSessionList(sess);

            return 0;
        }

        return -1;
    }

    return 0;
}


/*
 * SkipQuerySession - timeout process
 */
int SkipQuerySession(Session *sess, int icp_fd)
{
    Trace("SkipQuerySession: skip query, try to connect origin server\n");

    FD_CLR(icp_fd, &session_fdset);
    fd2sess[icp_fd].type = CONN_NOP;
    fd2sess[icp_fd].sess = NULL;

    release_socket(&icp_fd);

    CheckLastFd();

    if(!ConnectSession(sess)) {
        JoinSessionList(sess);
    }

    return 0;
}




/*
 * CheckSession - count number of session and relese very long session
 *
 *  return: status flag of sessions (like as chksum). It will use for identify.
 */
int CheckSessions()
{
    Session *s;
    int i;
    int nr, np, ni;
    int st;                     /* state of sessions */
    char diff_str[STRING_SIZE];
    time_t diff, life;
    int rem;


#if 0
    Trace("CheckSessions:\n");
#endif

    st = 0;
    nr = np = ni = 0;
    for(i=0;i<MAX_SESSION;i++) {
        s = &session_list[i];


        /*
         * remove strange session entry
         */
        if(s->type!=SESS_NOP && s->type!=SESS_CLIENT &&
            s->type!=SESS_PREFETCH && s->type!=SESS_IMPORT) {

            Log("; strange session (type %d): sess %p, %d th",
                s->type, s, i);

            LeaveSessionList(s);
            DeleteSession(s);
            continue;
        }

        /*
         * remove strange session entry
         */
        if(s->phase!=PHASE_UNKNOW && s->phase!=PHASE_IDLE &&
            s->phase!=PHASE_START && s->phase!=PHASE_REQUEST &&
            s->phase!=PHASE_THINK && s->phase!=PHASE_WAIT &&
            s->phase!=PHASE_SEND  &&
            s->phase!=PHASE_QUERY && s->phase!=PHASE_NAMERESOLV &&
            s->phase!=PHASE_CONNECTSIBLING && s->phase!=PHASE_CONNECTSERVER &&
            s->phase!=PHASE_HEADER && s->phase!=PHASE_BODY &&
            s->phase!=PHASE_END   && s->phase!=PHASE_ERROR) {

            Log("; strange session (phase %d): sess %p, %d th",
                s->phase, s, i);

            LeaveSessionList(s);
            DeleteSession(s);
            continue;
        }

        /*
         * skip idle session
         */
        if(s->phase==PHASE_IDLE) {
            continue;
        }

        if(s->phase==PHASE_WAIT) {
            ErrorAbortSession(s, HREP_INNERERROR,
                501, "Locked", "Sorry, such resource is locked");

            continue;
        }

        /*
         * remove error session
         */
        if(s->phase==PHASE_ERROR) {
            Log("; remove error session : sess %p, %d th : '%s'",
                s, i, s->_url);

            LeaveSessionList(s);
            DeleteSession(s);
            continue;
        }


        /*
         * remove old sessions
         */
        life =  WCOL_SUB_TIME(session_currenttime, s->begin);
        diff =  WCOL_SUB_TIME(session_currenttime, s->hot);

#if 0
        Trace("\t%p diff %d, life %d '%s'\n", s, diff, life, s->_url);
#endif

        rem  = 0;
        /* old request */
        if(s->phase==PHASE_REQUEST && diff > reception_timeout*1000) {
            Trace(
              "CheckSessions: give up long request session #%d, wait %d msec\n",
                i, diff);

            rem = 2;
        }
        /* cold connection */
        if(s->type==SESS_PREFETCH && diff > session_timeout) {
            Trace(
              "CheckSessions: give up cold long session #%d, wait %d msec\n",
                i, diff);

            rem = 3;
        }
        /* long connection */
        if(life > session_lifetime) {
            Trace(
              "CheckSessions: give up long session #%d, take %d msec\n",
                i, life);

            rem = 1;
        }

        if(rem) {
            strcpy(diff_str,
                display_msec(WCOL_SUB_TIME(session_currenttime, s->hot)));

            Trace("\t current         %10d.%06d (sec.usec)\n",
                session_currenttime.tv_sec, session_currenttime.tv_usec);
            Trace("\t hot             %10d.%06d (sec.usec)\n",
                s->hot.tv_sec, s->hot.tv_usec);
            Trace("\t diff            %10d.%03d    (sec.msec) [%s]\n",
                diff/1000, diff%1000, diff_str);
            Trace("\t reception limit %10d.%03d    (sec.msec)\n",
                reception_timeout, 0);
            Trace("\t session limit   %10d.%03d    (sec.msec)\n",
                session_timeout/1000, session_timeout%1000);

            if(rem==1) {
                ErrorAbortSession(s, HREP_IGNOREREQUEST,
                    408, "Connection Timeout",
                    "Sorry, cannot support long connection");
            }
            if(rem==2) {
                ErrorAbortSession(s, HREP_IGNOREREQUEST,
                    408, "Connection Timeout",
                    "Sorry, request is late. client down or busy");
            }
            if(rem==3) {
                ErrorAbortSession(s, HREP_IGNOREREQUEST,
                    408, "Connection Timeout",
                    "Sorry, response is late. server down or busy");
            }

            continue;
        }


        st += s->phase
                + s->hot.tv_usec
                + s->client_fd + s->server_fd + s->header_fd + s->body_fd;


        if(s->type==REQ_PREFETCH)   np++;
        if(s->type==REQ_CLIENT)     nr++;
        if(s->type==REQ_IMPORT)     ni++;


#if 0
        /*
         * Check 'How file descriptor are used ?'
         */
        if(s->client_fd != -1)  last++;
        if(s->server_fd != -1)  last++;
        if(s->header_fd != -1)  last++;
        if(s->body_fd   != -1)  last++;
#endif
    }
#if 0
    last_fd = last;
#endif

    num_session = nr + np + ni;
    num_client_session   = nr;
    num_prefetch_session = np;
    num_import_session   = ni;

    /*
    Trace("SESSION %d, FD %d\n", num_session, last_fd);
    */

    handlers[handler_no].busy = num_session;

    return st;
}




/*
 * state
 *          0   readable
 *          1   writeable
 *
*/

Session *find_session(fd2sess_t *fd2sess, int start, int end,
            fdset_t rdset, fd_set wdset,
            int *conntype, int *state, int *active_fd)
{
    int i;

#if 0
    Trace("time %d.%06d\n",
        session_currenttime.tv_sec, session_currenttime.tv_usec);
#endif
#if 0
    DisplaySessions(0);
#endif

    for(i=start;i<end;i++) {
        if(fd2sess[i].type==CONN_HTTP && FD_ISSET(i, &rdset)) {
            *state      = 0;
            *conntype   = CONN_HTTP;
            *active_fd  = i;

            FD_CLR(i, &rdset);
            return fd2sess[i].sess;
        }
        if(fd2sess[i].type==CONN_HTTP && FD_ISSET(i, &wdset)) {
            *state      = 1;
            *conntype   = CONN_HTTP;
            *active_fd  = i;

            FD_CLR(i, &wdset);
            return fd2sess[i].sess;
        }


        if(fd2sess[i].type==CONN_ICPQUERY && FD_ISSET(i, &rdset)) {
            *state = 0;
            *conntype = CONN_ICPREPLY;
            *active_fd = i;

            FD_CLR(i, &rdset);
            return fd2sess[i].sess;
        }


        if(fd2sess[i].type==CONN_ICPQUERY) {
            long kk;

            if(FD_ISSET(i, &wdset)) {
                *state = 1;
                *conntype = CONN_ICPREPLY;
                *active_fd = i;
                if(fd2sess[i].sess==NULL)
                    ClearQuerySession(*active_fd);

Trace("find_session: ICP reply arrive. fd# %d -> %p\n",
    i, fd2sess[i].sess);

                FD_CLR(i, &wdset);
                return fd2sess[i].sess;
            }

            kk = (long)WCOL_SUB_TIME(fd2sess[i].timeout, session_currenttime);
#if 0
            Trace("fd# %4d: type %x, time %d.%06d / %d.%06d -> %5d %c\n",
                i, fd2sess[i].type,
                fd2sess[i].timeout.tv_sec, fd2sess[i].timeout.tv_usec,
                session_currenttime.tv_sec, session_currenttime.tv_usec,
                kk, kk<=0L ? '-':'+'
            );
#endif

            if(kk<=0L) {
                *state = 1;
                *conntype = CONN_ICPTIMEOUT;
                *active_fd = i;
                if(fd2sess[i].sess==NULL)
                    ClearQuerySession(*active_fd);

Trace("find_session: ICP query timeout. fd# %d -> %p\n",
    i, fd2sess[i].sess);

                return fd2sess[i].sess;
            }
        }
    }

    return NULL;
}






/*
 * ScanActiveSession - scan active session
 *
 * returns:
 *      NULL    no input (or no wait)
 *              'state' specified writeable or readable
 *      !NULL   input
 *
 */
Session *ScanActiveSession(int servd, int command_fd, fdset_t session_fdset,
    time_t interval,
    int *conntype, int *state, int *active_fd)
{
    fdset_t         rdset, wdset, edset;
    struct timeval  timeout;
    int             select_chk;

    static int      cc=0;               /* count in loop */

    static int lpos=-1;
    Session *psess;
    int cpos, spos, epos;



#if 0
    Trace("ScanActiveSession: servd %d, command_fd %d\n", servd, command_fd);
#endif


    /*
     * Do select()
     */

    rdset = session_fdset;
    if(servd>=0) {
        FD_SET(servd, &rdset);
        if(servd>last_fd)
            last_fd = servd;
    }
    if(command_fd>=0) {
        FD_SET(command_fd, &rdset);
        if(command_fd>last_fd)
            last_fd = command_fd;
    }
    wdset = edset = rdset;

    timeout.tv_sec = interval; timeout.tv_usec = 0; /* wait several seconds */
    errno = 0;

    /*
     * when waiting requst and command, we should check readable
     */
    if(num_fd == 0) {
        select_chk = select(last_fd+1, &rdset, NULL, NULL, &timeout);
    }
    /*
     * other case, we should check readable and writeable
     */
    else {
        select_chk = select(last_fd+1, &rdset, &wdset, NULL, &timeout);
    }

#define chkfd(xx)   if(xx>=0 && isreadable(xx,1,0)==1) { \
                        Trace("\t%-10s %4d (readable)\n", #xx, xx); } \
                    else { \
                        Trace("\t%-10s %4d (unreadable)\n", #xx, xx); }

#if 0
    Trace("return of select() = %d / num_fd = %4d / last_fd = %4d\n",
        select_chk, num_fd, last_fd);
    chkfd(servd);
    chkfd(command_fd);
    dump_fdset(last_fd+1, &session_fdset, &rdset, NULL, NULL);
#endif

#undef chkfd

    /* Error */
    if(select_chk < 0) {
        if(errno == EINTR) {
            return (Session*) NULL;
        }
        if(errno == EBADF) {
#if 0
            dump_fdset(FD_SETSIZE, &session_fdset, &rdset, NULL, NULL);
#endif
            if(num_fd == 0) {
                    dump_fdset(last_fd+1, &edset, &rdset, NULL, NULL);
            }
            else {
                    dump_fdset(last_fd+1, &edset, &rdset, &wdset, NULL);
            }
        }

        Trace("ScanActiveSession: fail select()'s errno=%d\n", errno);
        Trace("last_fd %d, servd %d, command_fd %d\n",
            last_fd, servd, command_fd);

        return (Session*) NULL;
    }

    /* Timeout */
    if(select_chk == 0) {
        return (Session*) NULL;
    }

    /* reset */
    *conntype = CONN_ERROR;
    *state = -7;
    *active_fd = -1;


    /***
     *** FDSET was changed
     ***/
    cc++;
#if 0
    dump_fdset(FD_SETSIZE, &session_fdset,&rdset,&wdset,&edset);
#endif

#ifdef USE_ICMP
    /*
     * Scan ICMP
     */
    if(FD_ISSET(icmp_fd, &rdset) != 0) {
        ReadICMP(icmp_fd);
        return (Session*) NULL;
    }
#endif

    /*
     * Check New Connection
     */
    if(servd>=0 && FD_ISSET(servd, &rdset) != 0) {
        *conntype = CONN_HTTPREQUEST;

#if 0
        Trace("ScanActiveSession: servd(%d) is readable\n", servd);
#else
        Trace(">");
#endif
        *active_fd = servd;
        return (Session*) &session_fdset;   /* dummy */
    }

    /*
     * Check New Command
     */
    if(command_fd>=0 && FD_ISSET(command_fd, &rdset) != 0) {
        *conntype = CONN_COMMAND;

        Trace("ScanActiveSession: command_fd(%d) is readable\n", command_fd);
        *active_fd = command_fd;
        return (Session*) &session_fdset;   /* dummy */
    }

#if 0
    if( (command_fd<0 || (command_fd>=0 && last_fd<=command_fd))
        && ((servd<0 || (servd>=0 && last_fd<=servd))) ) {
        *conntype = CONN_NOP;
        return (Session*) NULL;
    }
#endif



#ifdef USE_FD_COMPACTION
    /*
     * Skip sessions
     */
/*
    if(servd == -1 && command_fd != -1 && command_fd == last_fd) {
*/
    if(last_fd == servd || last_fd == command_fd) {
#if 0
        Trace("ScanActiveSession: no job, skip session checking\n");
#endif

        *conntype = CONN_NOP;
        return (Session*) NULL;
    }

#endif /* USE_FD_COMPACTION */



    /*
     * Check Session
     */
    if(lpos<0)
        lpos = 0;
#if 0
    if(lpos>=FD_SETSIZE)
        lpos = 0;
#endif
    if(lpos>=last_fd+1)
        lpos = 0;


#ifdef TRACE_POS
    Trace("@w0 in  %d , %d \n", lpos, last_fd);
#endif

    if(lpos==0) {
        spos = 0;
#if 0
        epos = FD_SETSIZE;
#endif
        epos = last_fd+1;
        psess = find_session(fd2sess, spos, epos, rdset, wdset,
                conntype, state, active_fd);

#ifdef TRACE_POS
        Trace("@x_ a-fd# %d , %d [%d,%d]\n",
                *active_fd, lpos, spos, epos);
#endif
    }
    else {
        spos = lpos;
#if 0
        epos = FD_SETSIZE;
#endif
        epos = last_fd+1;
        psess = find_session(fd2sess, spos, epos, rdset, wdset,
                    conntype, state, active_fd);

#ifdef TRACE_POS
        Trace("@y1 a-fd# %d , %d [%d,%d]\n",
            *active_fd, lpos, spos, epos);
#endif

        if(psess==NULL) {
            spos = 0;
            epos = lpos;
            psess = find_session(fd2sess, spos, epos, rdset, wdset,
                    conntype, state, active_fd);

#ifdef TRACE_POS
            Trace("@y2 a-fd# %d , %d [%d,%d]\n",
                *active_fd, lpos, spos, epos);
#endif

        }
    }

    if(psess!=NULL) {
        lpos = *active_fd + 1;
#ifdef TRACE_POS
        Trace("@w1 out %d\n", lpos);
#endif

        return psess;
    }
    else {
        lpos = epos;

#ifdef TRACE_POS
        Trace("@w2 out %d\n", lpos);
#endif
    }


    /*
     * not found
     */

    *conntype = CONN_NOP;
    *active_fd = -1;
    return (Session*) NULL;
}


Session *ScanWaitSession()
{
    return (Session*) NULL;
}




char *_hrep_type_str[]= {"UNKNOW", "CACHEHIT", "CACHESAME", "SERVERACCESS",
            "SUCCESSPOST", "SERVERERROR", "IGNOREREQUEST", "INNERERROR",
            "INTRRUPT", "TIMEOUT" };


void ShowHReport(hreport_t *hrep)
{
    fprintf(stdout, "--- Handler Report ---\n");
    fprintf(stdout, "Type:       %s\n", HRepTypeStr(hrep->type));
    fprintf(stdout, "RequestId:  %d\n", hrep->id);
    fprintf(stdout, "Diff:       %d\n", (long)(hrep->diff));
    fprintf(stdout, "Keep-Alive: %s\n", hrep->keep_alive ? "YES" : "NO");
#if 0
    fprintf(stdout, "URL:        %s\n", hrep->url);
#endif
    ShowRequest(&hrep->req);
#if 0
    if(hrep->type==HREP_SERVERERROR) {
        fprintf(stdout, "State: %d\n", hrep->info.attr.state);
    }
    else
    if(hrep->type==HREP_CACHEHIT || hrep->type==HREP_SERVERACCESS
        || hrep->type==HREP_SUCCESSPOST) {

        if(trace_flag)
            What(&hrep->info);
    }
    else {
        fprintf(stdout, "\tand *junk*\n\n");
    }
#endif
    fflush(stdout);
}







static sigjmp_buf session_start;


/*
 *
 */
static RETSIGTYPE ProgramReset()
{
    Log("; SIGNAL QUIT ?");

    SIGLONGJMP(session_start, 1);
}


int n_access=0;
static int exit_flag=0;
static int add_toggle;


extern RETSIGTYPE ExitProgram();


int StatSess(Session* sess)
{
    int  olen;
    char a_type, a_err;

#if 0

    /*
     *
     */
if(sess->client_fd>0) {
    Trace("StatSess: client_fd %d\n", sess->client_fd);

    Trace("\tSOL_SOCKET/SO_TYPE  ");
    if(getsockopt(sess->client_fd, SOL_SOCKET, SO_TYPE , &a_type, &olen)<0) {
        Trace("error (%d)\n", errno);
    }
    else {
        Trace("value %d\n", a_type);
    }

    Trace("\tSOL_SOCKET/SO_ERROR ");
    if(getsockopt(sess->client_fd, SOL_SOCKET, SO_ERROR, &a_err, &olen)<0) {
        Trace("error (%d)\n", errno);
    }
    else {
        Trace("value %d\n", a_err);
    }
}

    /*
     *
     */
if(sess->server_fd>0) {
    Trace("StatSess: server_fd %d\n", sess->server_fd);

    Trace("\tSOL_SOCKET/SO_TYPE  ");
    if(getsockopt(sess->server_fd, SOL_SOCKET, SO_TYPE , &a_err, &olen)<0) {
        Trace("error (%d)\n", errno);
    }
    else {
        Trace("value %d\n", a_type);
    }

    Trace("\tSOL_SOCKET/SO_ERROR ");
    if(getsockopt(sess->server_fd, SOL_SOCKET, SO_ERROR, &a_err, &olen)<0) {
        Trace("error (%d)\n", errno);
    }
    else {
        Trace("value %d\n", a_type);
    }
}

#endif


    return 0;
}




extern int last_fd;
extern struct timeval session_currenttime;


extern int http_servd;

/*
 * main body of 'Handler'
 */
void Handler(int command_fd)
{
    Info    *info;
    mrequest_t  req;
    hreport_t   report;
    int     c, sum;
    int     fd = -1;
    int     ret;
    int     incache=0;
    extern  int handler_no;
    off_t   diff;
    char    memo[STRING_SIZE];
    int     keep_flag;
    int     count_in_conn;
    int     reg_sid_from_manager;

    int     chk, last_chk;
    int     c_chk;

    Session *sess, *nsess;
    int     conntype, state;
    int     intr_check_flag=0;

    int     active_fd;

    time_t  interval;

    /*
     * new log
     */
    wcol_pid = getpid();
    sprintf(memo, "handler %3d", handler_no);
    SetProcessComment(memo);

    if(trace_flag) {
        /*
        DLog("WAKE handler %d", handler_no);
        */

        SwitchLog(DEFAULT_DEBUGLOGDIR, "log of wcol's handler");
    }

#if 0
    global_ear = command_fd;
#endif
    fd = -1;

    interval = 10;


    /* handler must die, when manager was die */
    SetSignalHandler(SIGHUP,  ExitProgram);
/*
    SetSignalHandler(SIGTERM, ExitProgram);
*/
    SetSignalHandler(SIGTERM, HookTERM);
    SetSignalHandler(SIGQUIT, ProgramReset);

    info = (Info*)NULL;

    Trace("handler# %d\n", handler_no);

    request_seqno = &handlers[handler_no].z_req;
    *request_seqno = 0;
    session_seqno = &handlers[handler_no].z_sess;
    *session_seqno = 0;

    InitSessions(http_servd, command_fd);

    last_chk = -1;


    while(!catch_termination) {
        /*
         *
         */
        while(
          gettimeofday(&session_currenttime, NULL),
          CheckLastFd(),
          chk = CheckSessions(),
          LogStatus(),
          !catch_termination &&
          (sess=ScanActiveSession(http_servd, command_fd,
            session_fdset, interval, &conntype, &state, &active_fd))!=NULL) {

#if 0
            sess->hot = session_currenttime;
#endif


#if 0
            if( sess == &session_fdset ) {
                Trace(
    "Handler:-: fd# %3d, conntype %d, state %d\n",
                    active_fd, conntype, state);
            }
            else {
                Trace(
    "Handler:-: fd# %3d, conntype %d, state %d / sess %p, phase '%s' (%d)\n",
                    active_fd, conntype, state,
                    sess, SessPhaseStr(sess->phase), sess->phase);
            }
#endif
            switch(conntype) {
            case CONN_ERROR:
                Error("strange type event happen");
                break;

            case CONN_NOP:
                /* nothing */
                break;

            /* new command */
            case CONN_COMMAND:
                Trace("Handler: new command is arrive\n");
                c = RecvFRecord(command_fd, &fd, &req, sizeof(request_t));
				if(c==0) {
                    Trace("Handler: empty command. communication is over.\n");
					continue;
				}	
                if(c<0) {
                    Trace("Handler: RecvFRecord returns %d\n", c);
                    continue;
                }
                /*
                 * get current date/time
                 */
                gettimeofday(&session_currenttime, NULL);

                reg_sid_from_manager = req.sid;
#if 0
        Trace("Handler: RecvFRecord returns %d, fd %d, mode %s, sid %d\n",
                    c, fd, ReqModeStr(req.mode), reg_sid_from_manager);
                ShowRequest(&req);
#endif

                StartSession(NULL, fd, command_fd, &req);
#if 0
                if(fd!=-1) {
                    StartSession(NULL, fd, command_fd, &req);
                }
#endif


#ifdef MEMORY_CHECK
    show_memoryusage();
#endif

                break;

            /* new connection */
            case CONN_HTTPREQUEST:
#if 0
                Trace("Handler: new request is arrive\n");
#endif

                if(num_fd < max_session) {
                    if((nsess = NewSession())!=NULL) {
                        if((fd=AcceptRequest(http_servd))<0) {
                            /* ignore request or cannot lock port */
                            /*
                            Log("; ignored request (unacceptable)");
                            */
                        }
                        else {
                            StartSession(nsess, fd, command_fd, NULL);
                        }
                    }
                    else {
                        Log("; sorry, no new session space (%d/%d max %d)",
                                num_fd, most_num_session, max_session);
                    }
                }
                else {
                    Log("; sorry, session table is full (%d/%d max %d)",
                            num_session, most_num_session, max_session);
                }

#ifdef MEMORY_CHECK
    show_memoryusage();
#endif

                break;

            /* read/write */
            case CONN_HTTP:
                switch(sess->phase) {
                case PHASE_CONNECTSERVER:
#if 0
                    Trace("\n\tsid %d, phase CONNECTSERVER, state %d\n",
                        sess->req.sid, state);
#if 0
                    StatSess(sess);
#endif
                    ReachSession(sess);
#else
                    if(state==1) {                      /* WRITABLE */
                        ReachSession(sess);
                    }
                    if(state==0) {                      /* READABLE */
                        Trace("Handler: empty\n");
                        Log("; Connection fail '%s'", sess->_url);
                        ErrorAbortSession(sess, HREP_IGNOREREQUEST,
                            403, "Connect Error", "Server is down or busy");
                    }
#endif
                    break;
                case PHASE_HEADER:
                case PHASE_BODY:
                    if(state==0) {                      /* READABLE */
                        switch(ReadSession(sess)) {
                        case -1:                        /* Error */
                                AbortSession(sess);
                                break;
                        case  0:                        /* Success */
                                break;
                        case  1:                        /* End */
                                CloseSession(sess);
                                break;
                        default:
                                break;
                        }

                        while((sess=ScanWaitSession())!=NULL) {
                            LeaveWaitSessionList(sess);
                        }

                    }
                    break;
                default:
                    break;
                }
                break;

            /* receive query */
            case CONN_ICPQUERY:
                /* nothing */
                break;

            /* receive query's respone */
            case CONN_ICPREPLY:
#if 1
                Trace("Handler: sess %p (phase %d) with type %d, state %d, active_fd %d\n",
                    sess, sess->phase, conntype, state, active_fd);
#endif

                switch(sess->phase) {
                case PHASE_CONNECTSIBLING:
#if 1
                    if(state==1) {          /* writeable */
                        QuerySession(sess, active_fd);
                    }
#endif
                    break;
                case PHASE_QUERY:
                    if(state==0) {          /* readable */
                        ReceiveQueryReplySession(sess, active_fd);
                    }
                    if(state==1) {          /* writeable */
                        RemoveUnreachQuerySession(sess, active_fd);
                        /*
                        ReceiveQueryReplySession(sess, active_fd);
                        */
                    }
                    break;
                default:
                        RemoveUnreachQuerySession(sess, active_fd);
                }
                break;

            case CONN_ICPTIMEOUT:
#if 1
                Trace("Handler: conntype %d, state %d, active_fd %d\n",
                    conntype, state, active_fd);
#endif
                    SkipQuerySession(sess, active_fd);

                break;

            default:
                break;
            }
        }

    }


    if(catch_termination) {
        DLog("INFO exit normal");
    }
    else {
        DLog("INFO exit strange");
    }

    exit(WCOL_EXIT_NORMAL);

}



